sqlglot.dialects.presto API documentation

Edit on GitHub

# sqlglot.dialects.presto

View Source

      1from __future__ import annotations
      2
      3import typing as t
      4
      5from sqlglot import exp, generator, parser, tokens, transforms
      6from sqlglot.dialects.dialect import (
      7    Dialect,
      8    NormalizationStrategy,
      9    binary_from_function,
     10    bool_xor_sql,
     11    date_trunc_to_time,
     12    datestrtodate_sql,
     13    encode_decode_sql,
     14    build_formatted_time,
     15    if_sql,
     16    left_to_substring_sql,
     17    no_ilike_sql,
     18    no_pivot_sql,
     19    no_safe_divide_sql,
     20    no_timestamp_sql,
     21    regexp_extract_sql,
     22    rename_func,
     23    right_to_substring_sql,
     24    struct_extract_sql,
     25    str_position_sql,
     26    timestamptrunc_sql,
     27    timestrtotime_sql,
     28    ts_or_ds_add_cast,
     29    unit_to_str,
     30)
     31from sqlglot.dialects.hive import Hive
     32from sqlglot.dialects.mysql import MySQL
     33from sqlglot.helper import apply_index_offset, seq_get
     34from sqlglot.tokens import TokenType
     35from sqlglot.transforms import unqualify_columns
     36
     37
     38def _explode_to_unnest_sql(self: Presto.Generator, expression: exp.Lateral) -> str:
     39    if isinstance(expression.this, exp.Explode):
     40        return self.sql(
     41            exp.Join(
     42                this=exp.Unnest(
     43                    expressions=[expression.this.this],
     44                    alias=expression.args.get("alias"),
     45                    offset=isinstance(expression.this, exp.Posexplode),
     46                ),
     47                kind="cross",
     48            )
     49        )
     50    return self.lateral_sql(expression)
     51
     52
     53def _initcap_sql(self: Presto.Generator, expression: exp.Initcap) -> str:
     54    regex = r"(\w)(\w*)"
     55    return f"REGEXP_REPLACE({self.sql(expression, 'this')}, '{regex}', x -> UPPER(x[1]) || LOWER(x[2]))"
     56
     57
     58def _no_sort_array(self: Presto.Generator, expression: exp.SortArray) -> str:
     59    if expression.args.get("asc") == exp.false():
     60        comparator = "(a, b) -> CASE WHEN a < b THEN 1 WHEN a > b THEN -1 ELSE 0 END"
     61    else:
     62        comparator = None
     63    return self.func("ARRAY_SORT", expression.this, comparator)
     64
     65
     66def _schema_sql(self: Presto.Generator, expression: exp.Schema) -> str:
     67    if isinstance(expression.parent, exp.Property):
     68        columns = ", ".join(f"'{c.name}'" for c in expression.expressions)
     69        return f"ARRAY[{columns}]"
     70
     71    if expression.parent:
     72        for schema in expression.parent.find_all(exp.Schema):
     73            column_defs = schema.find_all(exp.ColumnDef)
     74            if column_defs and isinstance(schema.parent, exp.Property):
     75                expression.expressions.extend(column_defs)
     76
     77    return self.schema_sql(expression)
     78
     79
     80def _quantile_sql(self: Presto.Generator, expression: exp.Quantile) -> str:
     81    self.unsupported("Presto does not support exact quantiles")
     82    return self.func("APPROX_PERCENTILE", expression.this, expression.args.get("quantile"))
     83
     84
     85def _str_to_time_sql(
     86    self: Presto.Generator, expression: exp.StrToDate | exp.StrToTime | exp.TsOrDsToDate
     87) -> str:
     88    return self.func("DATE_PARSE", expression.this, self.format_time(expression))
     89
     90
     91def _ts_or_ds_to_date_sql(self: Presto.Generator, expression: exp.TsOrDsToDate) -> str:
     92    time_format = self.format_time(expression)
     93    if time_format and time_format not in (Presto.TIME_FORMAT, Presto.DATE_FORMAT):
     94        return self.sql(exp.cast(_str_to_time_sql(self, expression), exp.DataType.Type.DATE))
     95    return self.sql(
     96        exp.cast(exp.cast(expression.this, exp.DataType.Type.TIMESTAMP), exp.DataType.Type.DATE)
     97    )
     98
     99
    100def _ts_or_ds_add_sql(self: Presto.Generator, expression: exp.TsOrDsAdd) -> str:
    101    expression = ts_or_ds_add_cast(expression)
    102    unit = unit_to_str(expression)
    103    return self.func("DATE_ADD", unit, expression.expression, expression.this)
    104
    105
    106def _ts_or_ds_diff_sql(self: Presto.Generator, expression: exp.TsOrDsDiff) -> str:
    107    this = exp.cast(expression.this, exp.DataType.Type.TIMESTAMP)
    108    expr = exp.cast(expression.expression, exp.DataType.Type.TIMESTAMP)
    109    unit = unit_to_str(expression)
    110    return self.func("DATE_DIFF", unit, expr, this)
    111
    112
    113def _build_approx_percentile(args: t.List) -> exp.Expression:
    114    if len(args) == 4:
    115        return exp.ApproxQuantile(
    116            this=seq_get(args, 0),
    117            weight=seq_get(args, 1),
    118            quantile=seq_get(args, 2),
    119            accuracy=seq_get(args, 3),
    120        )
    121    if len(args) == 3:
    122        return exp.ApproxQuantile(
    123            this=seq_get(args, 0), quantile=seq_get(args, 1), accuracy=seq_get(args, 2)
    124        )
    125    return exp.ApproxQuantile.from_arg_list(args)
    126
    127
    128def _build_from_unixtime(args: t.List) -> exp.Expression:
    129    if len(args) == 3:
    130        return exp.UnixToTime(
    131            this=seq_get(args, 0),
    132            hours=seq_get(args, 1),
    133            minutes=seq_get(args, 2),
    134        )
    135    if len(args) == 2:
    136        return exp.UnixToTime(this=seq_get(args, 0), zone=seq_get(args, 1))
    137
    138    return exp.UnixToTime.from_arg_list(args)
    139
    140
    141def _unnest_sequence(expression: exp.Expression) -> exp.Expression:
    142    if isinstance(expression, exp.Table):
    143        if isinstance(expression.this, exp.GenerateSeries):
    144            unnest = exp.Unnest(expressions=[expression.this])
    145
    146            if expression.alias:
    147                return exp.alias_(unnest, alias="_u", table=[expression.alias], copy=False)
    148            return unnest
    149    return expression
    150
    151
    152def _first_last_sql(self: Presto.Generator, expression: exp.Func) -> str:
    153    """
    154    Trino doesn't support FIRST / LAST as functions, but they're valid in the context
    155    of MATCH_RECOGNIZE, so we need to preserve them in that case. In all other cases
    156    they're converted into an ARBITRARY call.
    157
    158    Reference: https://trino.io/docs/current/sql/match-recognize.html#logical-navigation-functions
    159    """
    160    if isinstance(expression.find_ancestor(exp.MatchRecognize, exp.Select), exp.MatchRecognize):
    161        return self.function_fallback_sql(expression)
    162
    163    return rename_func("ARBITRARY")(self, expression)
    164
    165
    166def _unix_to_time_sql(self: Presto.Generator, expression: exp.UnixToTime) -> str:
    167    scale = expression.args.get("scale")
    168    timestamp = self.sql(expression, "this")
    169    if scale in (None, exp.UnixToTime.SECONDS):
    170        return rename_func("FROM_UNIXTIME")(self, expression)
    171
    172    return f"FROM_UNIXTIME(CAST({timestamp} AS DOUBLE) / POW(10, {scale}))"
    173
    174
    175def _to_int(expression: exp.Expression) -> exp.Expression:
    176    if not expression.type:
    177        from sqlglot.optimizer.annotate_types import annotate_types
    178
    179        annotate_types(expression)
    180    if expression.type and expression.type.this not in exp.DataType.INTEGER_TYPES:
    181        return exp.cast(expression, to=exp.DataType.Type.BIGINT)
    182    return expression
    183
    184
    185def _build_to_char(args: t.List) -> exp.TimeToStr:
    186    fmt = seq_get(args, 1)
    187    if isinstance(fmt, exp.Literal):
    188        # We uppercase this to match Teradata's format mapping keys
    189        fmt.set("this", fmt.this.upper())
    190
    191    # We use "teradata" on purpose here, because the time formats are different in Presto.
    192    # See https://prestodb.io/docs/current/functions/teradata.html?highlight=to_char#to_char
    193    return build_formatted_time(exp.TimeToStr, "teradata")(args)
    194
    195
    196class Presto(Dialect):
    197    INDEX_OFFSET = 1
    198    NULL_ORDERING = "nulls_are_last"
    199    TIME_FORMAT = MySQL.TIME_FORMAT
    200    TIME_MAPPING = MySQL.TIME_MAPPING
    201    STRICT_STRING_CONCAT = True
    202    SUPPORTS_SEMI_ANTI_JOIN = False
    203    TYPED_DIVISION = True
    204    TABLESAMPLE_SIZE_IS_PERCENT = True
    205    LOG_BASE_FIRST: t.Optional[bool] = None
    206
    207    # https://github.com/trinodb/trino/issues/17
    208    # https://github.com/trinodb/trino/issues/12289
    209    # https://github.com/prestodb/presto/issues/2863
    210    NORMALIZATION_STRATEGY = NormalizationStrategy.CASE_INSENSITIVE
    211
    212    class Tokenizer(tokens.Tokenizer):
    213        UNICODE_STRINGS = [
    214            (prefix + q, q)
    215            for q in t.cast(t.List[str], tokens.Tokenizer.QUOTES)
    216            for prefix in ("U&", "u&")
    217        ]
    218
    219        KEYWORDS = {
    220            **tokens.Tokenizer.KEYWORDS,
    221            "START": TokenType.BEGIN,
    222            "MATCH_RECOGNIZE": TokenType.MATCH_RECOGNIZE,
    223            "ROW": TokenType.STRUCT,
    224            "IPADDRESS": TokenType.IPADDRESS,
    225            "IPPREFIX": TokenType.IPPREFIX,
    226            "TDIGEST": TokenType.TDIGEST,
    227            "HYPERLOGLOG": TokenType.HLLSKETCH,
    228        }
    229
    230        KEYWORDS.pop("QUALIFY")
    231
    232    class Parser(parser.Parser):
    233        VALUES_FOLLOWED_BY_PAREN = False
    234
    235        FUNCTIONS = {
    236            **parser.Parser.FUNCTIONS,
    237            "ARBITRARY": exp.AnyValue.from_arg_list,
    238            "APPROX_DISTINCT": exp.ApproxDistinct.from_arg_list,
    239            "APPROX_PERCENTILE": _build_approx_percentile,
    240            "BITWISE_AND": binary_from_function(exp.BitwiseAnd),
    241            "BITWISE_NOT": lambda args: exp.BitwiseNot(this=seq_get(args, 0)),
    242            "BITWISE_OR": binary_from_function(exp.BitwiseOr),
    243            "BITWISE_XOR": binary_from_function(exp.BitwiseXor),
    244            "CARDINALITY": exp.ArraySize.from_arg_list,
    245            "CONTAINS": exp.ArrayContains.from_arg_list,
    246            "DATE_ADD": lambda args: exp.DateAdd(
    247                this=seq_get(args, 2), expression=seq_get(args, 1), unit=seq_get(args, 0)
    248            ),
    249            "DATE_DIFF": lambda args: exp.DateDiff(
    250                this=seq_get(args, 2), expression=seq_get(args, 1), unit=seq_get(args, 0)
    251            ),
    252            "DATE_FORMAT": build_formatted_time(exp.TimeToStr, "presto"),
    253            "DATE_PARSE": build_formatted_time(exp.StrToTime, "presto"),
    254            "DATE_TRUNC": date_trunc_to_time,
    255            "ELEMENT_AT": lambda args: exp.Bracket(
    256                this=seq_get(args, 0), expressions=[seq_get(args, 1)], offset=1, safe=True
    257            ),
    258            "FROM_HEX": exp.Unhex.from_arg_list,
    259            "FROM_UNIXTIME": _build_from_unixtime,
    260            "FROM_UTF8": lambda args: exp.Decode(
    261                this=seq_get(args, 0), replace=seq_get(args, 1), charset=exp.Literal.string("utf-8")
    262            ),
    263            "NOW": exp.CurrentTimestamp.from_arg_list,
    264            "REGEXP_EXTRACT": lambda args: exp.RegexpExtract(
    265                this=seq_get(args, 0), expression=seq_get(args, 1), group=seq_get(args, 2)
    266            ),
    267            "REGEXP_REPLACE": lambda args: exp.RegexpReplace(
    268                this=seq_get(args, 0),
    269                expression=seq_get(args, 1),
    270                replacement=seq_get(args, 2) or exp.Literal.string(""),
    271            ),
    272            "ROW": exp.Struct.from_arg_list,
    273            "SEQUENCE": exp.GenerateSeries.from_arg_list,
    274            "SET_AGG": exp.ArrayUniqueAgg.from_arg_list,
    275            "SPLIT_TO_MAP": exp.StrToMap.from_arg_list,
    276            "STRPOS": lambda args: exp.StrPosition(
    277                this=seq_get(args, 0), substr=seq_get(args, 1), instance=seq_get(args, 2)
    278            ),
    279            "TO_CHAR": _build_to_char,
    280            "TO_UNIXTIME": exp.TimeToUnix.from_arg_list,
    281            "TO_UTF8": lambda args: exp.Encode(
    282                this=seq_get(args, 0), charset=exp.Literal.string("utf-8")
    283            ),
    284            "MD5": exp.MD5Digest.from_arg_list,
    285            "SHA256": lambda args: exp.SHA2(this=seq_get(args, 0), length=exp.Literal.number(256)),
    286            "SHA512": lambda args: exp.SHA2(this=seq_get(args, 0), length=exp.Literal.number(512)),
    287        }
    288
    289        FUNCTION_PARSERS = parser.Parser.FUNCTION_PARSERS.copy()
    290        FUNCTION_PARSERS.pop("TRIM")
    291
    292    class Generator(generator.Generator):
    293        INTERVAL_ALLOWS_PLURAL_FORM = False
    294        JOIN_HINTS = False
    295        TABLE_HINTS = False
    296        QUERY_HINTS = False
    297        IS_BOOL_ALLOWED = False
    298        TZ_TO_WITH_TIME_ZONE = True
    299        NVL2_SUPPORTED = False
    300        STRUCT_DELIMITER = ("(", ")")
    301        LIMIT_ONLY_LITERALS = True
    302        SUPPORTS_SINGLE_ARG_CONCAT = False
    303        LIKE_PROPERTY_INSIDE_SCHEMA = True
    304        MULTI_ARG_DISTINCT = False
    305        SUPPORTS_TO_NUMBER = False
    306        HEX_FUNC = "TO_HEX"
    307
    308        PROPERTIES_LOCATION = {
    309            **generator.Generator.PROPERTIES_LOCATION,
    310            exp.LocationProperty: exp.Properties.Location.UNSUPPORTED,
    311            exp.VolatileProperty: exp.Properties.Location.UNSUPPORTED,
    312        }
    313
    314        TYPE_MAPPING = {
    315            **generator.Generator.TYPE_MAPPING,
    316            exp.DataType.Type.INT: "INTEGER",
    317            exp.DataType.Type.FLOAT: "REAL",
    318            exp.DataType.Type.BINARY: "VARBINARY",
    319            exp.DataType.Type.TEXT: "VARCHAR",
    320            exp.DataType.Type.TIMETZ: "TIME",
    321            exp.DataType.Type.TIMESTAMPTZ: "TIMESTAMP",
    322            exp.DataType.Type.STRUCT: "ROW",
    323            exp.DataType.Type.DATETIME: "TIMESTAMP",
    324            exp.DataType.Type.DATETIME64: "TIMESTAMP",
    325            exp.DataType.Type.HLLSKETCH: "HYPERLOGLOG",
    326        }
    327
    328        TRANSFORMS = {
    329            **generator.Generator.TRANSFORMS,
    330            exp.AnyValue: rename_func("ARBITRARY"),
    331            exp.ApproxDistinct: lambda self, e: self.func(
    332                "APPROX_DISTINCT", e.this, e.args.get("accuracy")
    333            ),
    334            exp.ApproxQuantile: rename_func("APPROX_PERCENTILE"),
    335            exp.ArgMax: rename_func("MAX_BY"),
    336            exp.ArgMin: rename_func("MIN_BY"),
    337            exp.Array: lambda self, e: f"ARRAY[{self.expressions(e, flat=True)}]",
    338            exp.ArrayAny: rename_func("ANY_MATCH"),
    339            exp.ArrayConcat: rename_func("CONCAT"),
    340            exp.ArrayContains: rename_func("CONTAINS"),
    341            exp.ArraySize: rename_func("CARDINALITY"),
    342            exp.ArrayToString: rename_func("ARRAY_JOIN"),
    343            exp.ArrayUniqueAgg: rename_func("SET_AGG"),
    344            exp.AtTimeZone: rename_func("AT_TIMEZONE"),
    345            exp.BitwiseAnd: lambda self, e: self.func("BITWISE_AND", e.this, e.expression),
    346            exp.BitwiseLeftShift: lambda self, e: self.func(
    347                "BITWISE_ARITHMETIC_SHIFT_LEFT", e.this, e.expression
    348            ),
    349            exp.BitwiseNot: lambda self, e: self.func("BITWISE_NOT", e.this),
    350            exp.BitwiseOr: lambda self, e: self.func("BITWISE_OR", e.this, e.expression),
    351            exp.BitwiseRightShift: lambda self, e: self.func(
    352                "BITWISE_ARITHMETIC_SHIFT_RIGHT", e.this, e.expression
    353            ),
    354            exp.BitwiseXor: lambda self, e: self.func("BITWISE_XOR", e.this, e.expression),
    355            exp.Cast: transforms.preprocess([transforms.epoch_cast_to_ts]),
    356            exp.CurrentTimestamp: lambda *_: "CURRENT_TIMESTAMP",
    357            exp.DateAdd: lambda self, e: self.func(
    358                "DATE_ADD",
    359                unit_to_str(e),
    360                _to_int(e.expression),
    361                e.this,
    362            ),
    363            exp.DateDiff: lambda self, e: self.func(
    364                "DATE_DIFF", unit_to_str(e), e.expression, e.this
    365            ),
    366            exp.DateStrToDate: datestrtodate_sql,
    367            exp.DateToDi: lambda self,
    368            e: f"CAST(DATE_FORMAT({self.sql(e, 'this')}, {Presto.DATEINT_FORMAT}) AS INT)",
    369            exp.DateSub: lambda self, e: self.func(
    370                "DATE_ADD",
    371                unit_to_str(e),
    372                _to_int(e.expression * -1),
    373                e.this,
    374            ),
    375            exp.Decode: lambda self, e: encode_decode_sql(self, e, "FROM_UTF8"),
    376            exp.DiToDate: lambda self,
    377            e: f"CAST(DATE_PARSE(CAST({self.sql(e, 'this')} AS VARCHAR), {Presto.DATEINT_FORMAT}) AS DATE)",
    378            exp.Encode: lambda self, e: encode_decode_sql(self, e, "TO_UTF8"),
    379            exp.FileFormatProperty: lambda self, e: f"FORMAT='{e.name.upper()}'",
    380            exp.First: _first_last_sql,
    381            exp.FirstValue: _first_last_sql,
    382            exp.FromTimeZone: lambda self,
    383            e: f"WITH_TIMEZONE({self.sql(e, 'this')}, {self.sql(e, 'zone')}) AT TIME ZONE 'UTC'",
    384            exp.Group: transforms.preprocess([transforms.unalias_group]),
    385            exp.GroupConcat: lambda self, e: self.func(
    386                "ARRAY_JOIN", self.func("ARRAY_AGG", e.this), e.args.get("separator")
    387            ),
    388            exp.If: if_sql(),
    389            exp.ILike: no_ilike_sql,
    390            exp.Initcap: _initcap_sql,
    391            exp.ParseJSON: rename_func("JSON_PARSE"),
    392            exp.Last: _first_last_sql,
    393            exp.LastValue: _first_last_sql,
    394            exp.LastDay: lambda self, e: self.func("LAST_DAY_OF_MONTH", e.this),
    395            exp.Lateral: _explode_to_unnest_sql,
    396            exp.Left: left_to_substring_sql,
    397            exp.Levenshtein: rename_func("LEVENSHTEIN_DISTANCE"),
    398            exp.LogicalAnd: rename_func("BOOL_AND"),
    399            exp.LogicalOr: rename_func("BOOL_OR"),
    400            exp.Pivot: no_pivot_sql,
    401            exp.Quantile: _quantile_sql,
    402            exp.RegexpExtract: regexp_extract_sql,
    403            exp.Right: right_to_substring_sql,
    404            exp.SafeDivide: no_safe_divide_sql,
    405            exp.Schema: _schema_sql,
    406            exp.SchemaCommentProperty: lambda self, e: self.naked_property(e),
    407            exp.Select: transforms.preprocess(
    408                [
    409                    transforms.eliminate_qualify,
    410                    transforms.eliminate_distinct_on,
    411                    transforms.explode_to_unnest(1),
    412                    transforms.eliminate_semi_and_anti_joins,
    413                ]
    414            ),
    415            exp.SortArray: _no_sort_array,
    416            exp.StrPosition: lambda self, e: str_position_sql(self, e, generate_instance=True),
    417            exp.StrToDate: lambda self, e: f"CAST({_str_to_time_sql(self, e)} AS DATE)",
    418            exp.StrToMap: rename_func("SPLIT_TO_MAP"),
    419            exp.StrToTime: _str_to_time_sql,
    420            exp.StructExtract: struct_extract_sql,
    421            exp.Table: transforms.preprocess([_unnest_sequence]),
    422            exp.Timestamp: no_timestamp_sql,
    423            exp.TimestampTrunc: timestamptrunc_sql(),
    424            exp.TimeStrToDate: timestrtotime_sql,
    425            exp.TimeStrToTime: timestrtotime_sql,
    426            exp.TimeStrToUnix: lambda self, e: self.func(
    427                "TO_UNIXTIME", self.func("DATE_PARSE", e.this, Presto.TIME_FORMAT)
    428            ),
    429            exp.TimeToStr: lambda self, e: self.func("DATE_FORMAT", e.this, self.format_time(e)),
    430            exp.TimeToUnix: rename_func("TO_UNIXTIME"),
    431            exp.ToChar: lambda self, e: self.func("DATE_FORMAT", e.this, self.format_time(e)),
    432            exp.TryCast: transforms.preprocess([transforms.epoch_cast_to_ts]),
    433            exp.TsOrDiToDi: lambda self,
    434            e: f"CAST(SUBSTR(REPLACE(CAST({self.sql(e, 'this')} AS VARCHAR), '-', ''), 1, 8) AS INT)",
    435            exp.TsOrDsAdd: _ts_or_ds_add_sql,
    436            exp.TsOrDsDiff: _ts_or_ds_diff_sql,
    437            exp.TsOrDsToDate: _ts_or_ds_to_date_sql,
    438            exp.Unhex: rename_func("FROM_HEX"),
    439            exp.UnixToStr: lambda self,
    440            e: f"DATE_FORMAT(FROM_UNIXTIME({self.sql(e, 'this')}), {self.format_time(e)})",
    441            exp.UnixToTime: _unix_to_time_sql,
    442            exp.UnixToTimeStr: lambda self,
    443            e: f"CAST(FROM_UNIXTIME({self.sql(e, 'this')}) AS VARCHAR)",
    444            exp.VariancePop: rename_func("VAR_POP"),
    445            exp.With: transforms.preprocess([transforms.add_recursive_cte_column_names]),
    446            exp.WithinGroup: transforms.preprocess(
    447                [transforms.remove_within_group_for_percentiles]
    448            ),
    449            exp.Xor: bool_xor_sql,
    450            exp.MD5: lambda self, e: self.func(
    451                "LOWER", self.func("TO_HEX", self.func("MD5", self.sql(e, "this")))
    452            ),
    453            exp.MD5Digest: rename_func("MD5"),
    454            exp.SHA: rename_func("SHA1"),
    455            exp.SHA2: lambda self, e: self.func(
    456                "SHA256" if e.text("length") == "256" else "SHA512", e.this
    457            ),
    458        }
    459
    460        RESERVED_KEYWORDS = {
    461            "alter",
    462            "and",
    463            "as",
    464            "between",
    465            "by",
    466            "case",
    467            "cast",
    468            "constraint",
    469            "create",
    470            "cross",
    471            "current_time",
    472            "current_timestamp",
    473            "deallocate",
    474            "delete",
    475            "describe",
    476            "distinct",
    477            "drop",
    478            "else",
    479            "end",
    480            "escape",
    481            "except",
    482            "execute",
    483            "exists",
    484            "extract",
    485            "false",
    486            "for",
    487            "from",
    488            "full",
    489            "group",
    490            "having",
    491            "in",
    492            "inner",
    493            "insert",
    494            "intersect",
    495            "into",
    496            "is",
    497            "join",
    498            "left",
    499            "like",
    500            "natural",
    501            "not",
    502            "null",
    503            "on",
    504            "or",
    505            "order",
    506            "outer",
    507            "prepare",
    508            "right",
    509            "select",
    510            "table",
    511            "then",
    512            "true",
    513            "union",
    514            "using",
    515            "values",
    516            "when",
    517            "where",
    518            "with",
    519        }
    520
    521        def strtounix_sql(self, expression: exp.StrToUnix) -> str:
    522            # Since `TO_UNIXTIME` requires a `TIMESTAMP`, we need to parse the argument into one.
    523            # To do this, we first try to `DATE_PARSE` it, but since this can fail when there's a
    524            # timezone involved, we wrap it in a `TRY` call and use `PARSE_DATETIME` as a fallback,
    525            # which seems to be using the same time mapping as Hive, as per:
    526            # https://joda-time.sourceforge.net/apidocs/org/joda/time/format/DateTimeFormat.html
    527            value_as_text = exp.cast(expression.this, exp.DataType.Type.TEXT)
    528            parse_without_tz = self.func("DATE_PARSE", value_as_text, self.format_time(expression))
    529            parse_with_tz = self.func(
    530                "PARSE_DATETIME",
    531                value_as_text,
    532                self.format_time(expression, Hive.INVERSE_TIME_MAPPING, Hive.INVERSE_TIME_TRIE),
    533            )
    534            coalesced = self.func("COALESCE", self.func("TRY", parse_without_tz), parse_with_tz)
    535            return self.func("TO_UNIXTIME", coalesced)
    536
    537        def bracket_sql(self, expression: exp.Bracket) -> str:
    538            if expression.args.get("safe"):
    539                return self.func(
    540                    "ELEMENT_AT",
    541                    expression.this,
    542                    seq_get(
    543                        apply_index_offset(
    544                            expression.this,
    545                            expression.expressions,
    546                            1 - expression.args.get("offset", 0),
    547                        ),
    548                        0,
    549                    ),
    550                )
    551            return super().bracket_sql(expression)
    552
    553        def struct_sql(self, expression: exp.Struct) -> str:
    554            from sqlglot.optimizer.annotate_types import annotate_types
    555
    556            expression = annotate_types(expression)
    557            values: t.List[str] = []
    558            schema: t.List[str] = []
    559            unknown_type = False
    560
    561            for e in expression.expressions:
    562                if isinstance(e, exp.PropertyEQ):
    563                    if e.type and e.type.is_type(exp.DataType.Type.UNKNOWN):
    564                        unknown_type = True
    565                    else:
    566                        schema.append(f"{self.sql(e, 'this')} {self.sql(e.type)}")
    567                    values.append(self.sql(e, "expression"))
    568                else:
    569                    values.append(self.sql(e))
    570
    571            size = len(expression.expressions)
    572
    573            if not size or len(schema) != size:
    574                if unknown_type:
    575                    self.unsupported(
    576                        "Cannot convert untyped key-value definitions (try annotate_types)."
    577                    )
    578                return self.func("ROW", *values)
    579            return f"CAST(ROW({', '.join(values)}) AS ROW({', '.join(schema)}))"
    580
    581        def interval_sql(self, expression: exp.Interval) -> str:
    582            if expression.this and expression.text("unit").upper().startswith("WEEK"):
    583                return f"({expression.this.name} * INTERVAL '7' DAY)"
    584            return super().interval_sql(expression)
    585
    586        def transaction_sql(self, expression: exp.Transaction) -> str:
    587            modes = expression.args.get("modes")
    588            modes = f" {', '.join(modes)}" if modes else ""
    589            return f"START TRANSACTION{modes}"
    590
    591        def generateseries_sql(self, expression: exp.GenerateSeries) -> str:
    592            start = expression.args["start"]
    593            end = expression.args["end"]
    594            step = expression.args.get("step")
    595
    596            if isinstance(start, exp.Cast):
    597                target_type = start.to
    598            elif isinstance(end, exp.Cast):
    599                target_type = end.to
    600            else:
    601                target_type = None
    602
    603            if target_type and target_type.is_type("timestamp"):
    604                if target_type is start.to:
    605                    end = exp.cast(end, target_type)
    606                else:
    607                    start = exp.cast(start, target_type)
    608
    609            return self.func("SEQUENCE", start, end, step)
    610
    611        def offset_limit_modifiers(
    612            self, expression: exp.Expression, fetch: bool, limit: t.Optional[exp.Fetch | exp.Limit]
    613        ) -> t.List[str]:
    614            return [
    615                self.sql(expression, "offset"),
    616                self.sql(limit),
    617            ]
    618
    619        def create_sql(self, expression: exp.Create) -> str:
    620            """
    621            Presto doesn't support CREATE VIEW with expressions (ex: `CREATE VIEW x (cola)` then `(cola)` is the expression),
    622            so we need to remove them
    623            """
    624            kind = expression.args["kind"]
    625            schema = expression.this
    626            if kind == "VIEW" and schema.expressions:
    627                expression.this.set("expressions", None)
    628            return super().create_sql(expression)
    629
    630        def delete_sql(self, expression: exp.Delete) -> str:
    631            """
    632            Presto only supports DELETE FROM for a single table without an alias, so we need
    633            to remove the unnecessary parts. If the original DELETE statement contains more
    634            than one table to be deleted, we can't safely map it 1-1 to a Presto statement.
    635            """
    636            tables = expression.args.get("tables") or [expression.this]
    637            if len(tables) > 1:
    638                return super().delete_sql(expression)
    639
    640            table = tables[0]
    641            expression.set("this", table)
    642            expression.set("tables", None)
    643
    644            if isinstance(table, exp.Table):
    645                table_alias = table.args.get("alias")
    646                if table_alias:
    647                    table_alias.pop()
    648                    expression = t.cast(exp.Delete, expression.transform(unqualify_columns))
    649
    650            return super().delete_sql(expression)

class Presto(sqlglot.dialects.dialect.Dialect): View Source

    197class Presto(Dialect):
    198    INDEX_OFFSET = 1
    199    NULL_ORDERING = "nulls_are_last"
    200    TIME_FORMAT = MySQL.TIME_FORMAT
    201    TIME_MAPPING = MySQL.TIME_MAPPING
    202    STRICT_STRING_CONCAT = True
    203    SUPPORTS_SEMI_ANTI_JOIN = False
    204    TYPED_DIVISION = True
    205    TABLESAMPLE_SIZE_IS_PERCENT = True
    206    LOG_BASE_FIRST: t.Optional[bool] = None
    207
    208    # https://github.com/trinodb/trino/issues/17
    209    # https://github.com/trinodb/trino/issues/12289
    210    # https://github.com/prestodb/presto/issues/2863
    211    NORMALIZATION_STRATEGY = NormalizationStrategy.CASE_INSENSITIVE
    212
    213    class Tokenizer(tokens.Tokenizer):
    214        UNICODE_STRINGS = [
    215            (prefix + q, q)
    216            for q in t.cast(t.List[str], tokens.Tokenizer.QUOTES)
    217            for prefix in ("U&", "u&")
    218        ]
    219
    220        KEYWORDS = {
    221            **tokens.Tokenizer.KEYWORDS,
    222            "START": TokenType.BEGIN,
    223            "MATCH_RECOGNIZE": TokenType.MATCH_RECOGNIZE,
    224            "ROW": TokenType.STRUCT,
    225            "IPADDRESS": TokenType.IPADDRESS,
    226            "IPPREFIX": TokenType.IPPREFIX,
    227            "TDIGEST": TokenType.TDIGEST,
    228            "HYPERLOGLOG": TokenType.HLLSKETCH,
    229        }
    230
    231        KEYWORDS.pop("QUALIFY")
    232
    233    class Parser(parser.Parser):
    234        VALUES_FOLLOWED_BY_PAREN = False
    235
    236        FUNCTIONS = {
    237            **parser.Parser.FUNCTIONS,
    238            "ARBITRARY": exp.AnyValue.from_arg_list,
    239            "APPROX_DISTINCT": exp.ApproxDistinct.from_arg_list,
    240            "APPROX_PERCENTILE": _build_approx_percentile,
    241            "BITWISE_AND": binary_from_function(exp.BitwiseAnd),
    242            "BITWISE_NOT": lambda args: exp.BitwiseNot(this=seq_get(args, 0)),
    243            "BITWISE_OR": binary_from_function(exp.BitwiseOr),
    244            "BITWISE_XOR": binary_from_function(exp.BitwiseXor),
    245            "CARDINALITY": exp.ArraySize.from_arg_list,
    246            "CONTAINS": exp.ArrayContains.from_arg_list,
    247            "DATE_ADD": lambda args: exp.DateAdd(
    248                this=seq_get(args, 2), expression=seq_get(args, 1), unit=seq_get(args, 0)
    249            ),
    250            "DATE_DIFF": lambda args: exp.DateDiff(
    251                this=seq_get(args, 2), expression=seq_get(args, 1), unit=seq_get(args, 0)
    252            ),
    253            "DATE_FORMAT": build_formatted_time(exp.TimeToStr, "presto"),
    254            "DATE_PARSE": build_formatted_time(exp.StrToTime, "presto"),
    255            "DATE_TRUNC": date_trunc_to_time,
    256            "ELEMENT_AT": lambda args: exp.Bracket(
    257                this=seq_get(args, 0), expressions=[seq_get(args, 1)], offset=1, safe=True
    258            ),
    259            "FROM_HEX": exp.Unhex.from_arg_list,
    260            "FROM_UNIXTIME": _build_from_unixtime,
    261            "FROM_UTF8": lambda args: exp.Decode(
    262                this=seq_get(args, 0), replace=seq_get(args, 1), charset=exp.Literal.string("utf-8")
    263            ),
    264            "NOW": exp.CurrentTimestamp.from_arg_list,
    265            "REGEXP_EXTRACT": lambda args: exp.RegexpExtract(
    266                this=seq_get(args, 0), expression=seq_get(args, 1), group=seq_get(args, 2)
    267            ),
    268            "REGEXP_REPLACE": lambda args: exp.RegexpReplace(
    269                this=seq_get(args, 0),
    270                expression=seq_get(args, 1),
    271                replacement=seq_get(args, 2) or exp.Literal.string(""),
    272            ),
    273            "ROW": exp.Struct.from_arg_list,
    274            "SEQUENCE": exp.GenerateSeries.from_arg_list,
    275            "SET_AGG": exp.ArrayUniqueAgg.from_arg_list,
    276            "SPLIT_TO_MAP": exp.StrToMap.from_arg_list,
    277            "STRPOS": lambda args: exp.StrPosition(
    278                this=seq_get(args, 0), substr=seq_get(args, 1), instance=seq_get(args, 2)
    279            ),
    280            "TO_CHAR": _build_to_char,
    281            "TO_UNIXTIME": exp.TimeToUnix.from_arg_list,
    282            "TO_UTF8": lambda args: exp.Encode(
    283                this=seq_get(args, 0), charset=exp.Literal.string("utf-8")
    284            ),
    285            "MD5": exp.MD5Digest.from_arg_list,
    286            "SHA256": lambda args: exp.SHA2(this=seq_get(args, 0), length=exp.Literal.number(256)),
    287            "SHA512": lambda args: exp.SHA2(this=seq_get(args, 0), length=exp.Literal.number(512)),
    288        }
    289
    290        FUNCTION_PARSERS = parser.Parser.FUNCTION_PARSERS.copy()
    291        FUNCTION_PARSERS.pop("TRIM")
    292
    293    class Generator(generator.Generator):
    294        INTERVAL_ALLOWS_PLURAL_FORM = False
    295        JOIN_HINTS = False
    296        TABLE_HINTS = False
    297        QUERY_HINTS = False
    298        IS_BOOL_ALLOWED = False
    299        TZ_TO_WITH_TIME_ZONE = True
    300        NVL2_SUPPORTED = False
    301        STRUCT_DELIMITER = ("(", ")")
    302        LIMIT_ONLY_LITERALS = True
    303        SUPPORTS_SINGLE_ARG_CONCAT = False
    304        LIKE_PROPERTY_INSIDE_SCHEMA = True
    305        MULTI_ARG_DISTINCT = False
    306        SUPPORTS_TO_NUMBER = False
    307        HEX_FUNC = "TO_HEX"
    308
    309        PROPERTIES_LOCATION = {
    310            **generator.Generator.PROPERTIES_LOCATION,
    311            exp.LocationProperty: exp.Properties.Location.UNSUPPORTED,
    312            exp.VolatileProperty: exp.Properties.Location.UNSUPPORTED,
    313        }
    314
    315        TYPE_MAPPING = {
    316            **generator.Generator.TYPE_MAPPING,
    317            exp.DataType.Type.INT: "INTEGER",
    318            exp.DataType.Type.FLOAT: "REAL",
    319            exp.DataType.Type.BINARY: "VARBINARY",
    320            exp.DataType.Type.TEXT: "VARCHAR",
    321            exp.DataType.Type.TIMETZ: "TIME",
    322            exp.DataType.Type.TIMESTAMPTZ: "TIMESTAMP",
    323            exp.DataType.Type.STRUCT: "ROW",
    324            exp.DataType.Type.DATETIME: "TIMESTAMP",
    325            exp.DataType.Type.DATETIME64: "TIMESTAMP",
    326            exp.DataType.Type.HLLSKETCH: "HYPERLOGLOG",
    327        }
    328
    329        TRANSFORMS = {
    330            **generator.Generator.TRANSFORMS,
    331            exp.AnyValue: rename_func("ARBITRARY"),
    332            exp.ApproxDistinct: lambda self, e: self.func(
    333                "APPROX_DISTINCT", e.this, e.args.get("accuracy")
    334            ),
    335            exp.ApproxQuantile: rename_func("APPROX_PERCENTILE"),
    336            exp.ArgMax: rename_func("MAX_BY"),
    337            exp.ArgMin: rename_func("MIN_BY"),
    338            exp.Array: lambda self, e: f"ARRAY[{self.expressions(e, flat=True)}]",
    339            exp.ArrayAny: rename_func("ANY_MATCH"),
    340            exp.ArrayConcat: rename_func("CONCAT"),
    341            exp.ArrayContains: rename_func("CONTAINS"),
    342            exp.ArraySize: rename_func("CARDINALITY"),
    343            exp.ArrayToString: rename_func("ARRAY_JOIN"),
    344            exp.ArrayUniqueAgg: rename_func("SET_AGG"),
    345            exp.AtTimeZone: rename_func("AT_TIMEZONE"),
    346            exp.BitwiseAnd: lambda self, e: self.func("BITWISE_AND", e.this, e.expression),
    347            exp.BitwiseLeftShift: lambda self, e: self.func(
    348                "BITWISE_ARITHMETIC_SHIFT_LEFT", e.this, e.expression
    349            ),
    350            exp.BitwiseNot: lambda self, e: self.func("BITWISE_NOT", e.this),
    351            exp.BitwiseOr: lambda self, e: self.func("BITWISE_OR", e.this, e.expression),
    352            exp.BitwiseRightShift: lambda self, e: self.func(
    353                "BITWISE_ARITHMETIC_SHIFT_RIGHT", e.this, e.expression
    354            ),
    355            exp.BitwiseXor: lambda self, e: self.func("BITWISE_XOR", e.this, e.expression),
    356            exp.Cast: transforms.preprocess([transforms.epoch_cast_to_ts]),
    357            exp.CurrentTimestamp: lambda *_: "CURRENT_TIMESTAMP",
    358            exp.DateAdd: lambda self, e: self.func(
    359                "DATE_ADD",
    360                unit_to_str(e),
    361                _to_int(e.expression),
    362                e.this,
    363            ),
    364            exp.DateDiff: lambda self, e: self.func(
    365                "DATE_DIFF", unit_to_str(e), e.expression, e.this
    366            ),
    367            exp.DateStrToDate: datestrtodate_sql,
    368            exp.DateToDi: lambda self,
    369            e: f"CAST(DATE_FORMAT({self.sql(e, 'this')}, {Presto.DATEINT_FORMAT}) AS INT)",
    370            exp.DateSub: lambda self, e: self.func(
    371                "DATE_ADD",
    372                unit_to_str(e),
    373                _to_int(e.expression * -1),
    374                e.this,
    375            ),
    376            exp.Decode: lambda self, e: encode_decode_sql(self, e, "FROM_UTF8"),
    377            exp.DiToDate: lambda self,
    378            e: f"CAST(DATE_PARSE(CAST({self.sql(e, 'this')} AS VARCHAR), {Presto.DATEINT_FORMAT}) AS DATE)",
    379            exp.Encode: lambda self, e: encode_decode_sql(self, e, "TO_UTF8"),
    380            exp.FileFormatProperty: lambda self, e: f"FORMAT='{e.name.upper()}'",
    381            exp.First: _first_last_sql,
    382            exp.FirstValue: _first_last_sql,
    383            exp.FromTimeZone: lambda self,
    384            e: f"WITH_TIMEZONE({self.sql(e, 'this')}, {self.sql(e, 'zone')}) AT TIME ZONE 'UTC'",
    385            exp.Group: transforms.preprocess([transforms.unalias_group]),
    386            exp.GroupConcat: lambda self, e: self.func(
    387                "ARRAY_JOIN", self.func("ARRAY_AGG", e.this), e.args.get("separator")
    388            ),
    389            exp.If: if_sql(),
    390            exp.ILike: no_ilike_sql,
    391            exp.Initcap: _initcap_sql,
    392            exp.ParseJSON: rename_func("JSON_PARSE"),
    393            exp.Last: _first_last_sql,
    394            exp.LastValue: _first_last_sql,
    395            exp.LastDay: lambda self, e: self.func("LAST_DAY_OF_MONTH", e.this),
    396            exp.Lateral: _explode_to_unnest_sql,
    397            exp.Left: left_to_substring_sql,
    398            exp.Levenshtein: rename_func("LEVENSHTEIN_DISTANCE"),
    399            exp.LogicalAnd: rename_func("BOOL_AND"),
    400            exp.LogicalOr: rename_func("BOOL_OR"),
    401            exp.Pivot: no_pivot_sql,
    402            exp.Quantile: _quantile_sql,
    403            exp.RegexpExtract: regexp_extract_sql,
    404            exp.Right: right_to_substring_sql,
    405            exp.SafeDivide: no_safe_divide_sql,
    406            exp.Schema: _schema_sql,
    407            exp.SchemaCommentProperty: lambda self, e: self.naked_property(e),
    408            exp.Select: transforms.preprocess(
    409                [
    410                    transforms.eliminate_qualify,
    411                    transforms.eliminate_distinct_on,
    412                    transforms.explode_to_unnest(1),
    413                    transforms.eliminate_semi_and_anti_joins,
    414                ]
    415            ),
    416            exp.SortArray: _no_sort_array,
    417            exp.StrPosition: lambda self, e: str_position_sql(self, e, generate_instance=True),
    418            exp.StrToDate: lambda self, e: f"CAST({_str_to_time_sql(self, e)} AS DATE)",
    419            exp.StrToMap: rename_func("SPLIT_TO_MAP"),
    420            exp.StrToTime: _str_to_time_sql,
    421            exp.StructExtract: struct_extract_sql,
    422            exp.Table: transforms.preprocess([_unnest_sequence]),
    423            exp.Timestamp: no_timestamp_sql,
    424            exp.TimestampTrunc: timestamptrunc_sql(),
    425            exp.TimeStrToDate: timestrtotime_sql,
    426            exp.TimeStrToTime: timestrtotime_sql,
    427            exp.TimeStrToUnix: lambda self, e: self.func(
    428                "TO_UNIXTIME", self.func("DATE_PARSE", e.this, Presto.TIME_FORMAT)
    429            ),
    430            exp.TimeToStr: lambda self, e: self.func("DATE_FORMAT", e.this, self.format_time(e)),
    431            exp.TimeToUnix: rename_func("TO_UNIXTIME"),
    432            exp.ToChar: lambda self, e: self.func("DATE_FORMAT", e.this, self.format_time(e)),
    433            exp.TryCast: transforms.preprocess([transforms.epoch_cast_to_ts]),
    434            exp.TsOrDiToDi: lambda self,
    435            e: f"CAST(SUBSTR(REPLACE(CAST({self.sql(e, 'this')} AS VARCHAR), '-', ''), 1, 8) AS INT)",
    436            exp.TsOrDsAdd: _ts_or_ds_add_sql,
    437            exp.TsOrDsDiff: _ts_or_ds_diff_sql,
    438            exp.TsOrDsToDate: _ts_or_ds_to_date_sql,
    439            exp.Unhex: rename_func("FROM_HEX"),
    440            exp.UnixToStr: lambda self,
    441            e: f"DATE_FORMAT(FROM_UNIXTIME({self.sql(e, 'this')}), {self.format_time(e)})",
    442            exp.UnixToTime: _unix_to_time_sql,
    443            exp.UnixToTimeStr: lambda self,
    444            e: f"CAST(FROM_UNIXTIME({self.sql(e, 'this')}) AS VARCHAR)",
    445            exp.VariancePop: rename_func("VAR_POP"),
    446            exp.With: transforms.preprocess([transforms.add_recursive_cte_column_names]),
    447            exp.WithinGroup: transforms.preprocess(
    448                [transforms.remove_within_group_for_percentiles]
    449            ),
    450            exp.Xor: bool_xor_sql,
    451            exp.MD5: lambda self, e: self.func(
    452                "LOWER", self.func("TO_HEX", self.func("MD5", self.sql(e, "this")))
    453            ),
    454            exp.MD5Digest: rename_func("MD5"),
    455            exp.SHA: rename_func("SHA1"),
    456            exp.SHA2: lambda self, e: self.func(
    457                "SHA256" if e.text("length") == "256" else "SHA512", e.this
    458            ),
    459        }
    460
    461        RESERVED_KEYWORDS = {
    462            "alter",
    463            "and",
    464            "as",
    465            "between",
    466            "by",
    467            "case",
    468            "cast",
    469            "constraint",
    470            "create",
    471            "cross",
    472            "current_time",
    473            "current_timestamp",
    474            "deallocate",
    475            "delete",
    476            "describe",
    477            "distinct",
    478            "drop",
    479            "else",
    480            "end",
    481            "escape",
    482            "except",
    483            "execute",
    484            "exists",
    485            "extract",
    486            "false",
    487            "for",
    488            "from",
    489            "full",
    490            "group",
    491            "having",
    492            "in",
    493            "inner",
    494            "insert",
    495            "intersect",
    496            "into",
    497            "is",
    498            "join",
    499            "left",
    500            "like",
    501            "natural",
    502            "not",
    503            "null",
    504            "on",
    505            "or",
    506            "order",
    507            "outer",
    508            "prepare",
    509            "right",
    510            "select",
    511            "table",
    512            "then",
    513            "true",
    514            "union",
    515            "using",
    516            "values",
    517            "when",
    518            "where",
    519            "with",
    520        }
    521
    522        def strtounix_sql(self, expression: exp.StrToUnix) -> str:
    523            # Since `TO_UNIXTIME` requires a `TIMESTAMP`, we need to parse the argument into one.
    524            # To do this, we first try to `DATE_PARSE` it, but since this can fail when there's a
    525            # timezone involved, we wrap it in a `TRY` call and use `PARSE_DATETIME` as a fallback,
    526            # which seems to be using the same time mapping as Hive, as per:
    527            # https://joda-time.sourceforge.net/apidocs/org/joda/time/format/DateTimeFormat.html
    528            value_as_text = exp.cast(expression.this, exp.DataType.Type.TEXT)
    529            parse_without_tz = self.func("DATE_PARSE", value_as_text, self.format_time(expression))
    530            parse_with_tz = self.func(
    531                "PARSE_DATETIME",
    532                value_as_text,
    533                self.format_time(expression, Hive.INVERSE_TIME_MAPPING, Hive.INVERSE_TIME_TRIE),
    534            )
    535            coalesced = self.func("COALESCE", self.func("TRY", parse_without_tz), parse_with_tz)
    536            return self.func("TO_UNIXTIME", coalesced)
    537
    538        def bracket_sql(self, expression: exp.Bracket) -> str:
    539            if expression.args.get("safe"):
    540                return self.func(
    541                    "ELEMENT_AT",
    542                    expression.this,
    543                    seq_get(
    544                        apply_index_offset(
    545                            expression.this,
    546                            expression.expressions,
    547                            1 - expression.args.get("offset", 0),
    548                        ),
    549                        0,
    550                    ),
    551                )
    552            return super().bracket_sql(expression)
    553
    554        def struct_sql(self, expression: exp.Struct) -> str:
    555            from sqlglot.optimizer.annotate_types import annotate_types
    556
    557            expression = annotate_types(expression)
    558            values: t.List[str] = []
    559            schema: t.List[str] = []
    560            unknown_type = False
    561
    562            for e in expression.expressions:
    563                if isinstance(e, exp.PropertyEQ):
    564                    if e.type and e.type.is_type(exp.DataType.Type.UNKNOWN):
    565                        unknown_type = True
    566                    else:
    567                        schema.append(f"{self.sql(e, 'this')} {self.sql(e.type)}")
    568                    values.append(self.sql(e, "expression"))
    569                else:
    570                    values.append(self.sql(e))
    571
    572            size = len(expression.expressions)
    573
    574            if not size or len(schema) != size:
    575                if unknown_type:
    576                    self.unsupported(
    577                        "Cannot convert untyped key-value definitions (try annotate_types)."
    578                    )
    579                return self.func("ROW", *values)
    580            return f"CAST(ROW({', '.join(values)}) AS ROW({', '.join(schema)}))"
    581
    582        def interval_sql(self, expression: exp.Interval) -> str:
    583            if expression.this and expression.text("unit").upper().startswith("WEEK"):
    584                return f"({expression.this.name} * INTERVAL '7' DAY)"
    585            return super().interval_sql(expression)
    586
    587        def transaction_sql(self, expression: exp.Transaction) -> str:
    588            modes = expression.args.get("modes")
    589            modes = f" {', '.join(modes)}" if modes else ""
    590            return f"START TRANSACTION{modes}"
    591
    592        def generateseries_sql(self, expression: exp.GenerateSeries) -> str:
    593            start = expression.args["start"]
    594            end = expression.args["end"]
    595            step = expression.args.get("step")
    596
    597            if isinstance(start, exp.Cast):
    598                target_type = start.to
    599            elif isinstance(end, exp.Cast):
    600                target_type = end.to
    601            else:
    602                target_type = None
    603
    604            if target_type and target_type.is_type("timestamp"):
    605                if target_type is start.to:
    606                    end = exp.cast(end, target_type)
    607                else:
    608                    start = exp.cast(start, target_type)
    609
    610            return self.func("SEQUENCE", start, end, step)
    611
    612        def offset_limit_modifiers(
    613            self, expression: exp.Expression, fetch: bool, limit: t.Optional[exp.Fetch | exp.Limit]
    614        ) -> t.List[str]:
    615            return [
    616                self.sql(expression, "offset"),
    617                self.sql(limit),
    618            ]
    619
    620        def create_sql(self, expression: exp.Create) -> str:
    621            """
    622            Presto doesn't support CREATE VIEW with expressions (ex: `CREATE VIEW x (cola)` then `(cola)` is the expression),
    623            so we need to remove them
    624            """
    625            kind = expression.args["kind"]
    626            schema = expression.this
    627            if kind == "VIEW" and schema.expressions:
    628                expression.this.set("expressions", None)
    629            return super().create_sql(expression)
    630
    631        def delete_sql(self, expression: exp.Delete) -> str:
    632            """
    633            Presto only supports DELETE FROM for a single table without an alias, so we need
    634            to remove the unnecessary parts. If the original DELETE statement contains more
    635            than one table to be deleted, we can't safely map it 1-1 to a Presto statement.
    636            """
    637            tables = expression.args.get("tables") or [expression.this]
    638            if len(tables) > 1:
    639                return super().delete_sql(expression)
    640
    641            table = tables[0]
    642            expression.set("this", table)
    643            expression.set("tables", None)
    644
    645            if isinstance(table, exp.Table):
    646                table_alias = table.args.get("alias")
    647                if table_alias:
    648                    table_alias.pop()
    649                    expression = t.cast(exp.Delete, expression.transform(unqualify_columns))
    650
    651            return super().delete_sql(expression)

INDEX_OFFSET = 1

The base index offset for arrays.

NULL_ORDERING = 'nulls_are_last'

Default `NULL` ordering method to use if not explicitly set. Possible values: `"nulls_are_small"`, `"nulls_are_large"`, `"nulls_are_last"`

TIME_FORMAT = "'%Y-%m-%d %T'"
TIME_MAPPING = {'%M': '%B', '%c': '%-m', '%e': '%-d', '%h': '%I', '%i': '%M', '%s': '%S', '%u': '%W', '%k': '%-H', '%l': '%-I', '%T': '%H:%M:%S', '%W': '%a'}

Associates this dialect's time formats with their equivalent Python `strftime` formats.

STRICT_STRING_CONCAT = True

Whether `CONCAT`'s arguments must be strings.

SUPPORTS_SEMI_ANTI_JOIN = False

Whether `SEMI` or `ANTI` joins are supported.

TYPED_DIVISION = True

Whether the behavior of `a / b` depends on the types of `a` and `b`. False means `a / b` is always float division. True means `a / b` is integer division if both `a` and `b` are integers.

TABLESAMPLE_SIZE_IS_PERCENT = True

Whether a size in the table sample clause represents percentage.

LOG_BASE_FIRST: Optional[bool] = None

Whether the base comes first in the `LOG` function. Possible values: `True`, `False`, `None` (two arguments are not supported by `LOG`)

NORMALIZATION_STRATEGY = <NormalizationStrategy.CASE_INSENSITIVE: 'CASE_INSENSITIVE'>

Specifies the strategy according to which identifiers should be normalized.

tokenizer_class = <class 'Presto.Tokenizer'>
parser_class = <class 'Presto.Parser'>
generator_class = <class 'Presto.Generator'>
TIME_TRIE: Dict = {'%': {'M': {0: True}, 'c': {0: True}, 'e': {0: True}, 'h': {0: True}, 'i': {0: True}, 's': {0: True}, 'u': {0: True}, 'k': {0: True}, 'l': {0: True}, 'T': {0: True}, 'W': {0: True}}}
FORMAT_TRIE: Dict = {'%': {'M': {0: True}, 'c': {0: True}, 'e': {0: True}, 'h': {0: True}, 'i': {0: True}, 's': {0: True}, 'u': {0: True}, 'k': {0: True}, 'l': {0: True}, 'T': {0: True}, 'W': {0: True}}}
INVERSE_TIME_MAPPING: Dict[str, str] = {'%B': '%M', '%-m': '%c', '%-d': '%e', '%I': '%h', '%M': '%i', '%S': '%s', '%W': '%u', '%-H': '%k', '%-I': '%l', '%H:%M:%S': '%T', '%a': '%W'}
INVERSE_TIME_TRIE: Dict = {'%': {'B': {0: True}, '-': {'m': {0: True}, 'd': {0: True}, 'H': {0: True}, 'I': {0: True}}, 'I': {0: True}, 'M': {0: True}, 'S': {0: True}, 'W': {0: True}, 'H': {':': {'%': {'M': {':': {'%': {'S': {0: True}}}}}}}, 'a': {0: True}}}
ESCAPED_SEQUENCES: Dict[str, str] = {}
QUOTE_START = "'"
QUOTE_END = "'"
IDENTIFIER_START = '"'
IDENTIFIER_END = '"'
BIT_START: Optional[str] = None
BIT_END: Optional[str] = None
HEX_START: Optional[str] = None
HEX_END: Optional[str] = None
BYTE_START: Optional[str] = None
BYTE_END: Optional[str] = None
UNICODE_START: Optional[str] = "U&'"
UNICODE_END: Optional[str] = "'"

##### Inherited Members

sqlglot.dialects.dialect.Dialect
  Dialect
  WEEK_OFFSET
  UNNEST_COLUMN_ONLY
  ALIAS_POST_TABLESAMPLE
  IDENTIFIERS_CAN_START_WITH_DIGIT
  DPIPE_IS_STRING_CONCAT
  SUPPORTS_USER_DEFINED_TYPES
  NORMALIZE_FUNCTIONS
  SAFE_DIVISION
  CONCAT_COALESCE
  HEX_LOWERCASE
  DATE_FORMAT
  DATEINT_FORMAT
  FORMAT_MAPPING
  UNESCAPED_SEQUENCES
  PSEUDOCOLUMNS
  PREFER_CTE_ALIAS_COLUMN
  COPY_PARAMS_ARE_CSV
  get_or_raise
  format_time
  normalize_identifier
  case_sensitive
  can_identify
  quote_identifier
  to_json_path
  parse
  parse_into
  generate
  transpile
  tokenize
  tokenizer
  parser
  generator

class Presto.Tokenizer(sqlglot.tokens.Tokenizer): View Source

    213    class Tokenizer(tokens.Tokenizer):
    214        UNICODE_STRINGS = [
    215            (prefix + q, q)
    216            for q in t.cast(t.List[str], tokens.Tokenizer.QUOTES)
    217            for prefix in ("U&", "u&")
    218        ]
    219
    220        KEYWORDS = {
    221            **tokens.Tokenizer.KEYWORDS,
    222            "START": TokenType.BEGIN,
    223            "MATCH_RECOGNIZE": TokenType.MATCH_RECOGNIZE,
    224            "ROW": TokenType.STRUCT,
    225            "IPADDRESS": TokenType.IPADDRESS,
    226            "IPPREFIX": TokenType.IPPREFIX,
    227            "TDIGEST": TokenType.TDIGEST,
    228            "HYPERLOGLOG": TokenType.HLLSKETCH,
    229        }
    230
    231        KEYWORDS.pop("QUALIFY")

UNICODE_STRINGS = [("U&'", "'"), ("u&'", "'")]
KEYWORDS = {'{%': <TokenType.BLOCK_START: 'BLOCK_START'>, '{%+': <TokenType.BLOCK_START: 'BLOCK_START'>, '{%-': <TokenType.BLOCK_START: 'BLOCK_START'>, '%}': <TokenType.BLOCK_END: 'BLOCK_END'>, '+%}': <TokenType.BLOCK_END: 'BLOCK_END'>, '-%}': <TokenType.BLOCK_END: 'BLOCK_END'>, '{{+': <TokenType.BLOCK_START: 'BLOCK_START'>, '{{-': <TokenType.BLOCK_START: 'BLOCK_START'>, '+}}': <TokenType.BLOCK_END: 'BLOCK_END'>, '-}}': <TokenType.BLOCK_END: 'BLOCK_END'>, '/*+': <TokenType.HINT: 'HINT'>, '==': <TokenType.EQ: 'EQ'>, '::': <TokenType.DCOLON: 'DCOLON'>, '||': <TokenType.DPIPE: 'DPIPE'>, '>=': <TokenType.GTE: 'GTE'>, '<=': <TokenType.LTE: 'LTE'>, '<>': <TokenType.NEQ: 'NEQ'>, '!=': <TokenType.NEQ: 'NEQ'>, ':=': <TokenType.COLON_EQ: 'COLON_EQ'>, '<=>': <TokenType.NULLSAFE_EQ: 'NULLSAFE_EQ'>, '->': <TokenType.ARROW: 'ARROW'>, '->>': <TokenType.DARROW: 'DARROW'>, '=>': <TokenType.FARROW: 'FARROW'>, '#>': <TokenType.HASH_ARROW: 'HASH_ARROW'>, '#>>': <TokenType.DHASH_ARROW: 'DHASH_ARROW'>, '<->': <TokenType.LR_ARROW: 'LR_ARROW'>, '&&': <TokenType.DAMP: 'DAMP'>, '??': <TokenType.DQMARK: 'DQMARK'>, 'ALL': <TokenType.ALL: 'ALL'>, 'ALWAYS': <TokenType.ALWAYS: 'ALWAYS'>, 'AND': <TokenType.AND: 'AND'>, 'ANTI': <TokenType.ANTI: 'ANTI'>, 'ANY': <TokenType.ANY: 'ANY'>, 'ASC': <TokenType.ASC: 'ASC'>, 'AS': <TokenType.ALIAS: 'ALIAS'>, 'ASOF': <TokenType.ASOF: 'ASOF'>, 'AUTOINCREMENT': <TokenType.AUTO_INCREMENT: 'AUTO_INCREMENT'>, 'AUTO_INCREMENT': <TokenType.AUTO_INCREMENT: 'AUTO_INCREMENT'>, 'BEGIN': <TokenType.BEGIN: 'BEGIN'>, 'BETWEEN': <TokenType.BETWEEN: 'BETWEEN'>, 'CACHE': <TokenType.CACHE: 'CACHE'>, 'UNCACHE': <TokenType.UNCACHE: 'UNCACHE'>, 'CASE': <TokenType.CASE: 'CASE'>, 'CHARACTER SET': <TokenType.CHARACTER_SET: 'CHARACTER_SET'>, 'CLUSTER BY': <TokenType.CLUSTER_BY: 'CLUSTER_BY'>, 'COLLATE': <TokenType.COLLATE: 'COLLATE'>, 'COLUMN': <TokenType.COLUMN: 'COLUMN'>, 'COMMIT': <TokenType.COMMIT: 'COMMIT'>, 'CONNECT BY': <TokenType.CONNECT_BY: 'CONNECT_BY'>, 'CONSTRAINT': <TokenType.CONSTRAINT: 'CONSTRAINT'>, 'COPY': <TokenType.COPY: 'COPY'>, 'CREATE': <TokenType.CREATE: 'CREATE'>, 'CROSS': <TokenType.CROSS: 'CROSS'>, 'CUBE': <TokenType.CUBE: 'CUBE'>, 'CURRENT_DATE': <TokenType.CURRENT_DATE: 'CURRENT_DATE'>, 'CURRENT_TIME': <TokenType.CURRENT_TIME: 'CURRENT_TIME'>, 'CURRENT_TIMESTAMP': <TokenType.CURRENT_TIMESTAMP: 'CURRENT_TIMESTAMP'>, 'CURRENT_USER': <TokenType.CURRENT_USER: 'CURRENT_USER'>, 'DATABASE': <TokenType.DATABASE: 'DATABASE'>, 'DEFAULT': <TokenType.DEFAULT: 'DEFAULT'>, 'DELETE': <TokenType.DELETE: 'DELETE'>, 'DESC': <TokenType.DESC: 'DESC'>, 'DESCRIBE': <TokenType.DESCRIBE: 'DESCRIBE'>, 'DISTINCT': <TokenType.DISTINCT: 'DISTINCT'>, 'DISTRIBUTE BY': <TokenType.DISTRIBUTE_BY: 'DISTRIBUTE_BY'>, 'DIV': <TokenType.DIV: 'DIV'>, 'DROP': <TokenType.DROP: 'DROP'>, 'ELSE': <TokenType.ELSE: 'ELSE'>, 'END': <TokenType.END: 'END'>, 'ENUM': <TokenType.ENUM: 'ENUM'>, 'ESCAPE': <TokenType.ESCAPE: 'ESCAPE'>, 'EXCEPT': <TokenType.EXCEPT: 'EXCEPT'>, 'EXECUTE': <TokenType.EXECUTE: 'EXECUTE'>, 'EXISTS': <TokenType.EXISTS: 'EXISTS'>, 'FALSE': <TokenType.FALSE: 'FALSE'>, 'FETCH': <TokenType.FETCH: 'FETCH'>, 'FILTER': <TokenType.FILTER: 'FILTER'>, 'FIRST': <TokenType.FIRST: 'FIRST'>, 'FULL': <TokenType.FULL: 'FULL'>, 'FUNCTION': <TokenType.FUNCTION: 'FUNCTION'>, 'FOR': <TokenType.FOR: 'FOR'>, 'FOREIGN KEY': <TokenType.FOREIGN_KEY: 'FOREIGN_KEY'>, 'FORMAT': <TokenType.FORMAT: 'FORMAT'>, 'FROM': <TokenType.FROM: 'FROM'>, 'GEOGRAPHY': <TokenType.GEOGRAPHY: 'GEOGRAPHY'>, 'GEOMETRY': <TokenType.GEOMETRY: 'GEOMETRY'>, 'GLOB': <TokenType.GLOB: 'GLOB'>, 'GROUP BY': <TokenType.GROUP_BY: 'GROUP_BY'>, 'GROUPING SETS': <TokenType.GROUPING_SETS: 'GROUPING_SETS'>, 'HAVING': <TokenType.HAVING: 'HAVING'>, 'ILIKE': <TokenType.ILIKE: 'ILIKE'>, 'IN': <TokenType.IN: 'IN'>, 'INDEX': <TokenType.INDEX: 'INDEX'>, 'INET': <TokenType.INET: 'INET'>, 'INNER': <TokenType.INNER: 'INNER'>, 'INSERT': <TokenType.INSERT: 'INSERT'>, 'INTERVAL': <TokenType.INTERVAL: 'INTERVAL'>, 'INTERSECT': <TokenType.INTERSECT: 'INTERSECT'>, 'INTO': <TokenType.INTO: 'INTO'>, 'IS': <TokenType.IS: 'IS'>, 'ISNULL': <TokenType.ISNULL: 'ISNULL'>, 'JOIN': <TokenType.JOIN: 'JOIN'>, 'KEEP': <TokenType.KEEP: 'KEEP'>, 'KILL': <TokenType.KILL: 'KILL'>, 'LATERAL': <TokenType.LATERAL: 'LATERAL'>, 'LEFT': <TokenType.LEFT: 'LEFT'>, 'LIKE': <TokenType.LIKE: 'LIKE'>, 'LIMIT': <TokenType.LIMIT: 'LIMIT'>, 'LOAD': <TokenType.LOAD: 'LOAD'>, 'LOCK': <TokenType.LOCK: 'LOCK'>, 'MERGE': <TokenType.MERGE: 'MERGE'>, 'NATURAL': <TokenType.NATURAL: 'NATURAL'>, 'NEXT': <TokenType.NEXT: 'NEXT'>, 'NOT': <TokenType.NOT: 'NOT'>, 'NOTNULL': <TokenType.NOTNULL: 'NOTNULL'>, 'NULL': <TokenType.NULL: 'NULL'>, 'OBJECT': <TokenType.OBJECT: 'OBJECT'>, 'OFFSET': <TokenType.OFFSET: 'OFFSET'>, 'ON': <TokenType.ON: 'ON'>, 'OR': <TokenType.OR: 'OR'>, 'XOR': <TokenType.XOR: 'XOR'>, 'ORDER BY': <TokenType.ORDER_BY: 'ORDER_BY'>, 'ORDINALITY': <TokenType.ORDINALITY: 'ORDINALITY'>, 'OUTER': <TokenType.OUTER: 'OUTER'>, 'OVER': <TokenType.OVER: 'OVER'>, 'OVERLAPS': <TokenType.OVERLAPS: 'OVERLAPS'>, 'OVERWRITE': <TokenType.OVERWRITE: 'OVERWRITE'>, 'PARTITION': <TokenType.PARTITION: 'PARTITION'>, 'PARTITION BY': <TokenType.PARTITION_BY: 'PARTITION_BY'>, 'PARTITIONED BY': <TokenType.PARTITION_BY: 'PARTITION_BY'>, 'PARTITIONED_BY': <TokenType.PARTITION_BY: 'PARTITION_BY'>, 'PERCENT': <TokenType.PERCENT: 'PERCENT'>, 'PIVOT': <TokenType.PIVOT: 'PIVOT'>, 'PRAGMA': <TokenType.PRAGMA: 'PRAGMA'>, 'PRIMARY KEY': <TokenType.PRIMARY_KEY: 'PRIMARY_KEY'>, 'PROCEDURE': <TokenType.PROCEDURE: 'PROCEDURE'>, 'RANGE': <TokenType.RANGE: 'RANGE'>, 'RECURSIVE': <TokenType.RECURSIVE: 'RECURSIVE'>, 'REGEXP': <TokenType.RLIKE: 'RLIKE'>, 'REPLACE': <TokenType.REPLACE: 'REPLACE'>, 'RETURNING': <TokenType.RETURNING: 'RETURNING'>, 'REFERENCES': <TokenType.REFERENCES: 'REFERENCES'>, 'RIGHT': <TokenType.RIGHT: 'RIGHT'>, 'RLIKE': <TokenType.RLIKE: 'RLIKE'>, 'ROLLBACK': <TokenType.ROLLBACK: 'ROLLBACK'>, 'ROLLUP': <TokenType.ROLLUP: 'ROLLUP'>, 'ROW': <TokenType.STRUCT: 'STRUCT'>, 'ROWS': <TokenType.ROWS: 'ROWS'>, 'SCHEMA': <TokenType.SCHEMA: 'SCHEMA'>, 'SELECT': <TokenType.SELECT: 'SELECT'>, 'SEMI': <TokenType.SEMI: 'SEMI'>, 'SET': <TokenType.SET: 'SET'>, 'SETTINGS': <TokenType.SETTINGS: 'SETTINGS'>, 'SHOW': <TokenType.SHOW: 'SHOW'>, 'SIMILAR TO': <TokenType.SIMILAR_TO: 'SIMILAR_TO'>, 'SOME': <TokenType.SOME: 'SOME'>, 'SORT BY': <TokenType.SORT_BY: 'SORT_BY'>, 'START WITH': <TokenType.START_WITH: 'START_WITH'>, 'TABLE': <TokenType.TABLE: 'TABLE'>, 'TABLESAMPLE': <TokenType.TABLE_SAMPLE: 'TABLE_SAMPLE'>, 'TEMP': <TokenType.TEMPORARY: 'TEMPORARY'>, 'TEMPORARY': <TokenType.TEMPORARY: 'TEMPORARY'>, 'THEN': <TokenType.THEN: 'THEN'>, 'TRUE': <TokenType.TRUE: 'TRUE'>, 'TRUNCATE': <TokenType.TRUNCATE: 'TRUNCATE'>, 'UNION': <TokenType.UNION: 'UNION'>, 'UNKNOWN': <TokenType.UNKNOWN: 'UNKNOWN'>, 'UNNEST': <TokenType.UNNEST: 'UNNEST'>, 'UNPIVOT': <TokenType.UNPIVOT: 'UNPIVOT'>, 'UPDATE': <TokenType.UPDATE: 'UPDATE'>, 'USE': <TokenType.USE: 'USE'>, 'USING': <TokenType.USING: 'USING'>, 'UUID': <TokenType.UUID: 'UUID'>, 'VALUES': <TokenType.VALUES: 'VALUES'>, 'VIEW': <TokenType.VIEW: 'VIEW'>, 'VOLATILE': <TokenType.VOLATILE: 'VOLATILE'>, 'WHEN': <TokenType.WHEN: 'WHEN'>, 'WHERE': <TokenType.WHERE: 'WHERE'>, 'WINDOW': <TokenType.WINDOW: 'WINDOW'>, 'WITH': <TokenType.WITH: 'WITH'>, 'APPLY': <TokenType.APPLY: 'APPLY'>, 'ARRAY': <TokenType.ARRAY: 'ARRAY'>, 'BIT': <TokenType.BIT: 'BIT'>, 'BOOL': <TokenType.BOOLEAN: 'BOOLEAN'>, 'BOOLEAN': <TokenType.BOOLEAN: 'BOOLEAN'>, 'BYTE': <TokenType.TINYINT: 'TINYINT'>, 'MEDIUMINT': <TokenType.MEDIUMINT: 'MEDIUMINT'>, 'INT1': <TokenType.TINYINT: 'TINYINT'>, 'TINYINT': <TokenType.TINYINT: 'TINYINT'>, 'INT16': <TokenType.SMALLINT: 'SMALLINT'>, 'SHORT': <TokenType.SMALLINT: 'SMALLINT'>, 'SMALLINT': <TokenType.SMALLINT: 'SMALLINT'>, 'INT128': <TokenType.INT128: 'INT128'>, 'HUGEINT': <TokenType.INT128: 'INT128'>, 'INT2': <TokenType.SMALLINT: 'SMALLINT'>, 'INTEGER': <TokenType.INT: 'INT'>, 'INT': <TokenType.INT: 'INT'>, 'INT4': <TokenType.INT: 'INT'>, 'INT32': <TokenType.INT: 'INT'>, 'INT64': <TokenType.BIGINT: 'BIGINT'>, 'LONG': <TokenType.BIGINT: 'BIGINT'>, 'BIGINT': <TokenType.BIGINT: 'BIGINT'>, 'INT8': <TokenType.TINYINT: 'TINYINT'>, 'UINT': <TokenType.UINT: 'UINT'>, 'DEC': <TokenType.DECIMAL: 'DECIMAL'>, 'DECIMAL': <TokenType.DECIMAL: 'DECIMAL'>, 'BIGDECIMAL': <TokenType.BIGDECIMAL: 'BIGDECIMAL'>, 'BIGNUMERIC': <TokenType.BIGDECIMAL: 'BIGDECIMAL'>, 'MAP': <TokenType.MAP: 'MAP'>, 'NULLABLE': <TokenType.NULLABLE: 'NULLABLE'>, 'NUMBER': <TokenType.DECIMAL: 'DECIMAL'>, 'NUMERIC': <TokenType.DECIMAL: 'DECIMAL'>, 'FIXED': <TokenType.DECIMAL: 'DECIMAL'>, 'REAL': <TokenType.FLOAT: 'FLOAT'>, 'FLOAT': <TokenType.FLOAT: 'FLOAT'>, 'FLOAT4': <TokenType.FLOAT: 'FLOAT'>, 'FLOAT8': <TokenType.DOUBLE: 'DOUBLE'>, 'DOUBLE': <TokenType.DOUBLE: 'DOUBLE'>, 'DOUBLE PRECISION': <TokenType.DOUBLE: 'DOUBLE'>, 'JSON': <TokenType.JSON: 'JSON'>, 'CHAR': <TokenType.CHAR: 'CHAR'>, 'CHARACTER': <TokenType.CHAR: 'CHAR'>, 'NCHAR': <TokenType.NCHAR: 'NCHAR'>, 'VARCHAR': <TokenType.VARCHAR: 'VARCHAR'>, 'VARCHAR2': <TokenType.VARCHAR: 'VARCHAR'>, 'NVARCHAR': <TokenType.NVARCHAR: 'NVARCHAR'>, 'NVARCHAR2': <TokenType.NVARCHAR: 'NVARCHAR'>, 'BPCHAR': <TokenType.BPCHAR: 'BPCHAR'>, 'STR': <TokenType.TEXT: 'TEXT'>, 'STRING': <TokenType.TEXT: 'TEXT'>, 'TEXT': <TokenType.TEXT: 'TEXT'>, 'LONGTEXT': <TokenType.LONGTEXT: 'LONGTEXT'>, 'MEDIUMTEXT': <TokenType.MEDIUMTEXT: 'MEDIUMTEXT'>, 'TINYTEXT': <TokenType.TINYTEXT: 'TINYTEXT'>, 'CLOB': <TokenType.TEXT: 'TEXT'>, 'LONGVARCHAR': <TokenType.TEXT: 'TEXT'>, 'BINARY': <TokenType.BINARY: 'BINARY'>, 'BLOB': <TokenType.VARBINARY: 'VARBINARY'>, 'LONGBLOB': <TokenType.LONGBLOB: 'LONGBLOB'>, 'MEDIUMBLOB': <TokenType.MEDIUMBLOB: 'MEDIUMBLOB'>, 'TINYBLOB': <TokenType.TINYBLOB: 'TINYBLOB'>, 'BYTEA': <TokenType.VARBINARY: 'VARBINARY'>, 'VARBINARY': <TokenType.VARBINARY: 'VARBINARY'>, 'TIME': <TokenType.TIME: 'TIME'>, 'TIMETZ': <TokenType.TIMETZ: 'TIMETZ'>, 'TIMESTAMP': <TokenType.TIMESTAMP: 'TIMESTAMP'>, 'TIMESTAMPTZ': <TokenType.TIMESTAMPTZ: 'TIMESTAMPTZ'>, 'TIMESTAMPLTZ': <TokenType.TIMESTAMPLTZ: 'TIMESTAMPLTZ'>, 'TIMESTAMP_LTZ': <TokenType.TIMESTAMPLTZ: 'TIMESTAMPLTZ'>, 'TIMESTAMPNTZ': <TokenType.TIMESTAMPNTZ: 'TIMESTAMPNTZ'>, 'TIMESTAMP_NTZ': <TokenType.TIMESTAMPNTZ: 'TIMESTAMPNTZ'>, 'DATE': <TokenType.DATE: 'DATE'>, 'DATETIME': <TokenType.DATETIME: 'DATETIME'>, 'INT4RANGE': <TokenType.INT4RANGE: 'INT4RANGE'>, 'INT4MULTIRANGE': <TokenType.INT4MULTIRANGE: 'INT4MULTIRANGE'>, 'INT8RANGE': <TokenType.INT8RANGE: 'INT8RANGE'>, 'INT8MULTIRANGE': <TokenType.INT8MULTIRANGE: 'INT8MULTIRANGE'>, 'NUMRANGE': <TokenType.NUMRANGE: 'NUMRANGE'>, 'NUMMULTIRANGE': <TokenType.NUMMULTIRANGE: 'NUMMULTIRANGE'>, 'TSRANGE': <TokenType.TSRANGE: 'TSRANGE'>, 'TSMULTIRANGE': <TokenType.TSMULTIRANGE: 'TSMULTIRANGE'>, 'TSTZRANGE': <TokenType.TSTZRANGE: 'TSTZRANGE'>, 'TSTZMULTIRANGE': <TokenType.TSTZMULTIRANGE: 'TSTZMULTIRANGE'>, 'DATERANGE': <TokenType.DATERANGE: 'DATERANGE'>, 'DATEMULTIRANGE': <TokenType.DATEMULTIRANGE: 'DATEMULTIRANGE'>, 'UNIQUE': <TokenType.UNIQUE: 'UNIQUE'>, 'STRUCT': <TokenType.STRUCT: 'STRUCT'>, 'SEQUENCE': <TokenType.SEQUENCE: 'SEQUENCE'>, 'VARIANT': <TokenType.VARIANT: 'VARIANT'>, 'ALTER': <TokenType.ALTER: 'ALTER'>, 'ANALYZE': <TokenType.COMMAND: 'COMMAND'>, 'CALL': <TokenType.COMMAND: 'COMMAND'>, 'COMMENT': <TokenType.COMMENT: 'COMMENT'>, 'EXPLAIN': <TokenType.COMMAND: 'COMMAND'>, 'GRANT': <TokenType.COMMAND: 'COMMAND'>, 'OPTIMIZE': <TokenType.COMMAND: 'COMMAND'>, 'PREPARE': <TokenType.COMMAND: 'COMMAND'>, 'VACUUM': <TokenType.COMMAND: 'COMMAND'>, 'USER-DEFINED': <TokenType.USERDEFINED: 'USERDEFINED'>, 'FOR VERSION': <TokenType.VERSION_SNAPSHOT: 'VERSION_SNAPSHOT'>, 'FOR TIMESTAMP': <TokenType.TIMESTAMP_SNAPSHOT: 'TIMESTAMP_SNAPSHOT'>, 'START': <TokenType.BEGIN: 'BEGIN'>, 'MATCH_RECOGNIZE': <TokenType.MATCH_RECOGNIZE: 'MATCH_RECOGNIZE'>, 'IPADDRESS': <TokenType.IPADDRESS: 'IPADDRESS'>, 'IPPREFIX': <TokenType.IPPREFIX: 'IPPREFIX'>, 'TDIGEST': <TokenType.TDIGEST: 'TDIGEST'>, 'HYPERLOGLOG': <TokenType.HLLSKETCH: 'HLLSKETCH'>}

##### Inherited Members

sqlglot.tokens.Tokenizer
  Tokenizer
  SINGLE_TOKENS
  BIT_STRINGS
  BYTE_STRINGS
  HEX_STRINGS
  RAW_STRINGS
  HEREDOC_STRINGS
  IDENTIFIERS
  IDENTIFIER_ESCAPES
  QUOTES
  STRING_ESCAPES
  VAR_SINGLE_TOKENS
  HEREDOC_TAG_IS_IDENTIFIER
  HEREDOC_STRING_ALTERNATIVE
  WHITE_SPACE
  COMMANDS
  COMMAND_PREFIX_TOKENS
  NUMERIC_LITERALS
  COMMENTS
  dialect
  reset
  tokenize
  tokenize_rs
  size
  sql
  tokens

class Presto.Parser(sqlglot.parser.Parser): View Source

    233    class Parser(parser.Parser):
    234        VALUES_FOLLOWED_BY_PAREN = False
    235
    236        FUNCTIONS = {
    237            **parser.Parser.FUNCTIONS,
    238            "ARBITRARY": exp.AnyValue.from_arg_list,
    239            "APPROX_DISTINCT": exp.ApproxDistinct.from_arg_list,
    240            "APPROX_PERCENTILE": _build_approx_percentile,
    241            "BITWISE_AND": binary_from_function(exp.BitwiseAnd),
    242            "BITWISE_NOT": lambda args: exp.BitwiseNot(this=seq_get(args, 0)),
    243            "BITWISE_OR": binary_from_function(exp.BitwiseOr),
    244            "BITWISE_XOR": binary_from_function(exp.BitwiseXor),
    245            "CARDINALITY": exp.ArraySize.from_arg_list,
    246            "CONTAINS": exp.ArrayContains.from_arg_list,
    247            "DATE_ADD": lambda args: exp.DateAdd(
    248                this=seq_get(args, 2), expression=seq_get(args, 1), unit=seq_get(args, 0)
    249            ),
    250            "DATE_DIFF": lambda args: exp.DateDiff(
    251                this=seq_get(args, 2), expression=seq_get(args, 1), unit=seq_get(args, 0)
    252            ),
    253            "DATE_FORMAT": build_formatted_time(exp.TimeToStr, "presto"),
    254            "DATE_PARSE": build_formatted_time(exp.StrToTime, "presto"),
    255            "DATE_TRUNC": date_trunc_to_time,
    256            "ELEMENT_AT": lambda args: exp.Bracket(
    257                this=seq_get(args, 0), expressions=[seq_get(args, 1)], offset=1, safe=True
    258            ),
    259            "FROM_HEX": exp.Unhex.from_arg_list,
    260            "FROM_UNIXTIME": _build_from_unixtime,
    261            "FROM_UTF8": lambda args: exp.Decode(
    262                this=seq_get(args, 0), replace=seq_get(args, 1), charset=exp.Literal.string("utf-8")
    263            ),
    264            "NOW": exp.CurrentTimestamp.from_arg_list,
    265            "REGEXP_EXTRACT": lambda args: exp.RegexpExtract(
    266                this=seq_get(args, 0), expression=seq_get(args, 1), group=seq_get(args, 2)
    267            ),
    268            "REGEXP_REPLACE": lambda args: exp.RegexpReplace(
    269                this=seq_get(args, 0),
    270                expression=seq_get(args, 1),
    271                replacement=seq_get(args, 2) or exp.Literal.string(""),
    272            ),
    273            "ROW": exp.Struct.from_arg_list,
    274            "SEQUENCE": exp.GenerateSeries.from_arg_list,
    275            "SET_AGG": exp.ArrayUniqueAgg.from_arg_list,
    276            "SPLIT_TO_MAP": exp.StrToMap.from_arg_list,
    277            "STRPOS": lambda args: exp.StrPosition(
    278                this=seq_get(args, 0), substr=seq_get(args, 1), instance=seq_get(args, 2)
    279            ),
    280            "TO_CHAR": _build_to_char,
    281            "TO_UNIXTIME": exp.TimeToUnix.from_arg_list,
    282            "TO_UTF8": lambda args: exp.Encode(
    283                this=seq_get(args, 0), charset=exp.Literal.string("utf-8")
    284            ),
    285            "MD5": exp.MD5Digest.from_arg_list,
    286            "SHA256": lambda args: exp.SHA2(this=seq_get(args, 0), length=exp.Literal.number(256)),
    287            "SHA512": lambda args: exp.SHA2(this=seq_get(args, 0), length=exp.Literal.number(512)),
    288        }
    289
    290        FUNCTION_PARSERS = parser.Parser.FUNCTION_PARSERS.copy()
    291        FUNCTION_PARSERS.pop("TRIM")

Parser consumes a list of tokens produced by the Tokenizer and produces a parsed syntax tree.

###### Arguments:

- error_level: The desired error level. Default: ErrorLevel.IMMEDIATE
- error_message_context: The amount of context to capture from a query string when displaying the error message (in number of characters). Default: 100
- max_errors: Maximum number of error messages to include in a raised ParseError. This is only relevant if error_level is ErrorLevel.RAISE. Default: 3

VALUES_FOLLOWED_BY_PAREN = False
FUNCTIONS = {'ABS': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Abs'>>, 'ADD_MONTHS': <bound method Func.from_arg_list of <class 'sqlglot.expressions.AddMonths'>>, 'ANONYMOUS_AGG_FUNC': <bound method Func.from_arg_list of <class 'sqlglot.expressions.AnonymousAggFunc'>>, 'ANY_VALUE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.AnyValue'>>, 'APPROX_DISTINCT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ApproxDistinct'>>, 'APPROX_COUNT_DISTINCT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ApproxDistinct'>>, 'APPROX_QUANTILE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ApproxQuantile'>>, 'APPROX_TOP_K': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ApproxTopK'>>, 'ARG_MAX': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArgMax'>>, 'ARGMAX': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArgMax'>>, 'MAX_BY': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArgMax'>>, 'ARG_MIN': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArgMin'>>, 'ARGMIN': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArgMin'>>, 'MIN_BY': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArgMin'>>, 'ARRAY': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Array'>>, 'ARRAY_AGG': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArrayAgg'>>, 'ARRAY_ALL': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArrayAll'>>, 'ARRAY_ANY': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArrayAny'>>, 'ARRAY_CONCAT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArrayConcat'>>, 'ARRAY_CAT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArrayConcat'>>, 'ARRAY_CONSTRUCT_COMPACT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArrayConstructCompact'>>, 'ARRAY_CONTAINS': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArrayContains'>>, 'FILTER': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArrayFilter'>>, 'ARRAY_FILTER': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArrayFilter'>>, 'ARRAY_OVERLAPS': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArrayOverlaps'>>, 'ARRAY_SIZE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArraySize'>>, 'ARRAY_LENGTH': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArraySize'>>, 'ARRAY_SORT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArraySort'>>, 'ARRAY_SUM': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArraySum'>>, 'ARRAY_TO_STRING': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArrayToString'>>, 'ARRAY_JOIN': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArrayToString'>>, 'ARRAY_UNION_AGG': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArrayUnionAgg'>>, 'ARRAY_UNIQUE_AGG': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArrayUniqueAgg'>>, 'AVG': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Avg'>>, 'CASE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Case'>>, 'CAST': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Cast'>>, 'CAST_TO_STR_TYPE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.CastToStrType'>>, 'CBRT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Cbrt'>>, 'CEIL': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Ceil'>>, 'CEILING': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Ceil'>>, 'CHR': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Chr'>>, 'CHAR': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Chr'>>, 'COALESCE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Coalesce'>>, 'IFNULL': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Coalesce'>>, 'NVL': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Coalesce'>>, 'COLLATE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Collate'>>, 'COMBINED_AGG_FUNC': <bound method Func.from_arg_list of <class 'sqlglot.expressions.CombinedAggFunc'>>, 'COMBINED_PARAMETERIZED_AGG': <bound method Func.from_arg_list of <class 'sqlglot.expressions.CombinedParameterizedAgg'>>, 'CONCAT': <function Parser.<lambda>>, 'CONCAT_WS': <function Parser.<lambda>>, 'CONNECT_BY_ROOT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ConnectByRoot'>>, 'CONVERT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Convert'>>, 'CORR': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Corr'>>, 'COUNT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Count'>>, 'COUNT_IF': <bound method Func.from_arg_list of <class 'sqlglot.expressions.CountIf'>>, 'COUNTIF': <bound method Func.from_arg_list of <class 'sqlglot.expressions.CountIf'>>, 'COVAR_POP': <bound method Func.from_arg_list of <class 'sqlglot.expressions.CovarPop'>>, 'COVAR_SAMP': <bound method Func.from_arg_list of <class 'sqlglot.expressions.CovarSamp'>>, 'CURRENT_DATE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.CurrentDate'>>, 'CURRENT_DATETIME': <bound method Func.from_arg_list of <class 'sqlglot.expressions.CurrentDatetime'>>, 'CURRENT_TIME': <bound method Func.from_arg_list of <class 'sqlglot.expressions.CurrentTime'>>, 'CURRENT_TIMESTAMP': <bound method Func.from_arg_list of <class 'sqlglot.expressions.CurrentTimestamp'>>, 'CURRENT_USER': <bound method Func.from_arg_list of <class 'sqlglot.expressions.CurrentUser'>>, 'DATE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Date'>>, 'DATE_ADD': <function Presto.Parser.<lambda>>, 'DATEDIFF': <bound method Func.from_arg_list of <class 'sqlglot.expressions.DateDiff'>>, 'DATE_DIFF': <function Presto.Parser.<lambda>>, 'DATE_FROM_PARTS': <bound method Func.from_arg_list of <class 'sqlglot.expressions.DateFromParts'>>, 'DATEFROMPARTS': <bound method Func.from_arg_list of <class 'sqlglot.expressions.DateFromParts'>>, 'DATE_STR_TO_DATE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.DateStrToDate'>>, 'DATE_SUB': <bound method Func.from_arg_list of <class 'sqlglot.expressions.DateSub'>>, 'DATE_TO_DATE_STR': <function Parser.<lambda>>, 'DATE_TO_DI': <bound method Func.from_arg_list of <class 'sqlglot.expressions.DateToDi'>>, 'DATE_TRUNC': <function date_trunc_to_time>, 'DATETIME_ADD': <bound method Func.from_arg_list of <class 'sqlglot.expressions.DatetimeAdd'>>, 'DATETIME_DIFF': <bound method Func.from_arg_list of <class 'sqlglot.expressions.DatetimeDiff'>>, 'DATETIME_SUB': <bound method Func.from_arg_list of <class 'sqlglot.expressions.DatetimeSub'>>, 'DATETIME_TRUNC': <bound method Func.from_arg_list of <class 'sqlglot.expressions.DatetimeTrunc'>>, 'DAY': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Day'>>, 'DAY_OF_MONTH': <bound method Func.from_arg_list of <class 'sqlglot.expressions.DayOfMonth'>>, 'DAYOFMONTH': <bound method Func.from_arg_list of <class 'sqlglot.expressions.DayOfMonth'>>, 'DAY_OF_WEEK': <bound method Func.from_arg_list of <class 'sqlglot.expressions.DayOfWeek'>>, 'DAYOFWEEK': <bound method Func.from_arg_list of <class 'sqlglot.expressions.DayOfWeek'>>, 'DAY_OF_YEAR': <bound method Func.from_arg_list of <class 'sqlglot.expressions.DayOfYear'>>, 'DAYOFYEAR': <bound method Func.from_arg_list of <class 'sqlglot.expressions.DayOfYear'>>, 'DECODE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Decode'>>, 'DI_TO_DATE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.DiToDate'>>, 'ENCODE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Encode'>>, 'EXP': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Exp'>>, 'EXPLODE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Explode'>>, 'EXPLODE_OUTER': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ExplodeOuter'>>, 'EXTRACT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Extract'>>, 'FIRST': <bound method Func.from_arg_list of <class 'sqlglot.expressions.First'>>, 'FIRST_VALUE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.FirstValue'>>, 'FLATTEN': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Flatten'>>, 'FLOOR': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Floor'>>, 'FROM_BASE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.FromBase'>>, 'FROM_BASE64': <bound method Func.from_arg_list of <class 'sqlglot.expressions.FromBase64'>>, 'GENERATE_DATE_ARRAY': <bound method Func.from_arg_list of <class 'sqlglot.expressions.GenerateDateArray'>>, 'GENERATE_SERIES': <bound method Func.from_arg_list of <class 'sqlglot.expressions.GenerateSeries'>>, 'GREATEST': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Greatest'>>, 'GROUP_CONCAT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.GroupConcat'>>, 'HEX': <function build_hex>, 'HLL': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Hll'>>, 'IF': <bound method Func.from_arg_list of <class 'sqlglot.expressions.If'>>, 'IIF': <bound method Func.from_arg_list of <class 'sqlglot.expressions.If'>>, 'INITCAP': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Initcap'>>, 'IS_INF': <bound method Func.from_arg_list of <class 'sqlglot.expressions.IsInf'>>, 'ISINF': <bound method Func.from_arg_list of <class 'sqlglot.expressions.IsInf'>>, 'IS_NAN': <bound method Func.from_arg_list of <class 'sqlglot.expressions.IsNan'>>, 'ISNAN': <bound method Func.from_arg_list of <class 'sqlglot.expressions.IsNan'>>, 'J_S_O_N_ARRAY': <bound method Func.from_arg_list of <class 'sqlglot.expressions.JSONArray'>>, 'J_S_O_N_ARRAY_AGG': <bound method Func.from_arg_list of <class 'sqlglot.expressions.JSONArrayAgg'>>, 'JSON_ARRAY_CONTAINS': <bound method Func.from_arg_list of <class 'sqlglot.expressions.JSONArrayContains'>>, 'JSONB_EXTRACT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.JSONBExtract'>>, 'JSONB_EXTRACT_SCALAR': <bound method Func.from_arg_list of <class 'sqlglot.expressions.JSONBExtractScalar'>>, 'JSON_EXTRACT': <function build_extract_json_with_path.<locals>._builder>, 'JSON_EXTRACT_SCALAR': <function build_extract_json_with_path.<locals>._builder>, 'JSON_FORMAT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.JSONFormat'>>, 'J_S_O_N_OBJECT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.JSONObject'>>, 'J_S_O_N_OBJECT_AGG': <bound method Func.from_arg_list of <class 'sqlglot.expressions.JSONObjectAgg'>>, 'J_S_O_N_TABLE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.JSONTable'>>, 'LAG': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Lag'>>, 'LAST': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Last'>>, 'LAST_DAY': <bound method Func.from_arg_list of <class 'sqlglot.expressions.LastDay'>>, 'LAST_DAY_OF_MONTH': <bound method Func.from_arg_list of <class 'sqlglot.expressions.LastDay'>>, 'LAST_VALUE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.LastValue'>>, 'LEAD': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Lead'>>, 'LEAST': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Least'>>, 'LEFT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Left'>>, 'LENGTH': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Length'>>, 'LEN': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Length'>>, 'LEVENSHTEIN': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Levenshtein'>>, 'LN': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Ln'>>, 'LOG': <function build_logarithm>, 'LOGICAL_AND': <bound method Func.from_arg_list of <class 'sqlglot.expressions.LogicalAnd'>>, 'BOOL_AND': <bound method Func.from_arg_list of <class 'sqlglot.expressions.LogicalAnd'>>, 'BOOLAND_AGG': <bound method Func.from_arg_list of <class 'sqlglot.expressions.LogicalAnd'>>, 'LOGICAL_OR': <bound method Func.from_arg_list of <class 'sqlglot.expressions.LogicalOr'>>, 'BOOL_OR': <bound method Func.from_arg_list of <class 'sqlglot.expressions.LogicalOr'>>, 'BOOLOR_AGG': <bound method Func.from_arg_list of <class 'sqlglot.expressions.LogicalOr'>>, 'LOWER': <function build_lower>, 'LCASE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Lower'>>, 'LOWER_HEX': <bound method Func.from_arg_list of <class 'sqlglot.expressions.LowerHex'>>, 'MD5': <bound method Func.from_arg_list of <class 'sqlglot.expressions.MD5Digest'>>, 'MD5_DIGEST': <bound method Func.from_arg_list of <class 'sqlglot.expressions.MD5Digest'>>, 'MAP': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Map'>>, 'MAP_FROM_ENTRIES': <bound method Func.from_arg_list of <class 'sqlglot.expressions.MapFromEntries'>>, 'MATCH_AGAINST': <bound method Func.from_arg_list of <class 'sqlglot.expressions.MatchAgainst'>>, 'MAX': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Max'>>, 'MIN': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Min'>>, 'MONTH': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Month'>>, 'MONTHS_BETWEEN': <bound method Func.from_arg_list of <class 'sqlglot.expressions.MonthsBetween'>>, 'NEXT_VALUE_FOR': <bound method Func.from_arg_list of <class 'sqlglot.expressions.NextValueFor'>>, 'NTH_VALUE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.NthValue'>>, 'NULLIF': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Nullif'>>, 'NUMBER_TO_STR': <bound method Func.from_arg_list of <class 'sqlglot.expressions.NumberToStr'>>, 'NVL2': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Nvl2'>>, 'OPEN_J_S_O_N': <bound method Func.from_arg_list of <class 'sqlglot.expressions.OpenJSON'>>, 'PARAMETERIZED_AGG': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ParameterizedAgg'>>, 'PARSE_JSON': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ParseJSON'>>, 'JSON_PARSE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ParseJSON'>>, 'PERCENTILE_CONT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.PercentileCont'>>, 'PERCENTILE_DISC': <bound method Func.from_arg_list of <class 'sqlglot.expressions.PercentileDisc'>>, 'POSEXPLODE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Posexplode'>>, 'POSEXPLODE_OUTER': <bound method Func.from_arg_list of <class 'sqlglot.expressions.PosexplodeOuter'>>, 'POWER': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Pow'>>, 'POW': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Pow'>>, 'PREDICT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Predict'>>, 'QUANTILE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Quantile'>>, 'QUARTER': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Quarter'>>, 'RAND': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Rand'>>, 'RANDOM': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Rand'>>, 'RANDN': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Randn'>>, 'RANGE_N': <bound method Func.from_arg_list of <class 'sqlglot.expressions.RangeN'>>, 'READ_CSV': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ReadCSV'>>, 'REDUCE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Reduce'>>, 'REGEXP_EXTRACT': <function Presto.Parser.<lambda>>, 'REGEXP_I_LIKE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.RegexpILike'>>, 'REGEXP_LIKE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.RegexpLike'>>, 'REGEXP_REPLACE': <function Presto.Parser.<lambda>>, 'REGEXP_SPLIT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.RegexpSplit'>>, 'REPEAT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Repeat'>>, 'RIGHT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Right'>>, 'ROUND': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Round'>>, 'ROW_NUMBER': <bound method Func.from_arg_list of <class 'sqlglot.expressions.RowNumber'>>, 'SHA': <bound method Func.from_arg_list of <class 'sqlglot.expressions.SHA'>>, 'SHA1': <bound method Func.from_arg_list of <class 'sqlglot.expressions.SHA'>>, 'SHA2': <bound method Func.from_arg_list of <class 'sqlglot.expressions.SHA2'>>, 'SAFE_DIVIDE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.SafeDivide'>>, 'SIGN': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Sign'>>, 'SIGNUM': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Sign'>>, 'SORT_ARRAY': <bound method Func.from_arg_list of <class 'sqlglot.expressions.SortArray'>>, 'SPLIT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Split'>>, 'SQRT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Sqrt'>>, 'STANDARD_HASH': <bound method Func.from_arg_list of <class 'sqlglot.expressions.StandardHash'>>, 'STAR_MAP': <bound method Func.from_arg_list of <class 'sqlglot.expressions.StarMap'>>, 'STARTS_WITH': <bound method Func.from_arg_list of <class 'sqlglot.expressions.StartsWith'>>, 'STARTSWITH': <bound method Func.from_arg_list of <class 'sqlglot.expressions.StartsWith'>>, 'STDDEV': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Stddev'>>, 'STDDEV_POP': <bound method Func.from_arg_list of <class 'sqlglot.expressions.StddevPop'>>, 'STDDEV_SAMP': <bound method Func.from_arg_list of <class 'sqlglot.expressions.StddevSamp'>>, 'STR_POSITION': <bound method Func.from_arg_list of <class 'sqlglot.expressions.StrPosition'>>, 'STR_TO_DATE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.StrToDate'>>, 'STR_TO_MAP': <bound method Func.from_arg_list of <class 'sqlglot.expressions.StrToMap'>>, 'STR_TO_TIME': <bound method Func.from_arg_list of <class 'sqlglot.expressions.StrToTime'>>, 'STR_TO_UNIX': <bound method Func.from_arg_list of <class 'sqlglot.expressions.StrToUnix'>>, 'STRUCT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Struct'>>, 'STRUCT_EXTRACT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.StructExtract'>>, 'STUFF': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Stuff'>>, 'INSERT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Stuff'>>, 'SUBSTRING': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Substring'>>, 'SUM': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Sum'>>, 'TIME_ADD': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TimeAdd'>>, 'TIME_DIFF': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TimeDiff'>>, 'TIME_FROM_PARTS': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TimeFromParts'>>, 'TIMEFROMPARTS': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TimeFromParts'>>, 'TIME_STR_TO_DATE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TimeStrToDate'>>, 'TIME_STR_TO_TIME': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TimeStrToTime'>>, 'TIME_STR_TO_UNIX': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TimeStrToUnix'>>, 'TIME_SUB': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TimeSub'>>, 'TIME_TO_STR': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TimeToStr'>>, 'TIME_TO_TIME_STR': <function Parser.<lambda>>, 'TIME_TO_UNIX': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TimeToUnix'>>, 'TIME_TRUNC': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TimeTrunc'>>, 'TIMESTAMP': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Timestamp'>>, 'TIMESTAMP_ADD': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TimestampAdd'>>, 'TIMESTAMPDIFF': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TimestampDiff'>>, 'TIMESTAMP_DIFF': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TimestampDiff'>>, 'TIMESTAMP_FROM_PARTS': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TimestampFromParts'>>, 'TIMESTAMPFROMPARTS': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TimestampFromParts'>>, 'TIMESTAMP_SUB': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TimestampSub'>>, 'TIMESTAMP_TRUNC': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TimestampTrunc'>>, 'TO_ARRAY': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ToArray'>>, 'TO_BASE64': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ToBase64'>>, 'TO_CHAR': <function _build_to_char>, 'TO_DAYS': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ToDays'>>, 'TO_MAP': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ToMap'>>, 'TO_NUMBER': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ToNumber'>>, 'TRANSFORM': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Transform'>>, 'TRIM': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Trim'>>, 'TRY': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Try'>>, 'TRY_CAST': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TryCast'>>, 'TS_OR_DI_TO_DI': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TsOrDiToDi'>>, 'TS_OR_DS_ADD': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TsOrDsAdd'>>, 'TS_OR_DS_DIFF': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TsOrDsDiff'>>, 'TS_OR_DS_TO_DATE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TsOrDsToDate'>>, 'TS_OR_DS_TO_DATE_STR': <function Parser.<lambda>>, 'TS_OR_DS_TO_TIME': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TsOrDsToTime'>>, 'TS_OR_DS_TO_TIMESTAMP': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TsOrDsToTimestamp'>>, 'UNHEX': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Unhex'>>, 'UNIX_DATE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.UnixDate'>>, 'UNIX_TO_STR': <bound method Func.from_arg_list of <class 'sqlglot.expressions.UnixToStr'>>, 'UNIX_TO_TIME': <bound method Func.from_arg_list of <class 'sqlglot.expressions.UnixToTime'>>, 'UNIX_TO_TIME_STR': <bound method Func.from_arg_list of <class 'sqlglot.expressions.UnixToTimeStr'>>, 'UPPER': <function build_upper>, 'UCASE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Upper'>>, 'VAR_MAP': <function build_var_map>, 'VARIANCE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Variance'>>, 'VARIANCE_SAMP': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Variance'>>, 'VAR_SAMP': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Variance'>>, 'VARIANCE_POP': <bound method Func.from_arg_list of <class 'sqlglot.expressions.VariancePop'>>, 'VAR_POP': <bound method Func.from_arg_list of <class 'sqlglot.expressions.VariancePop'>>, 'WEEK': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Week'>>, 'WEEK_OF_YEAR': <bound method Func.from_arg_list of <class 'sqlglot.expressions.WeekOfYear'>>, 'WEEKOFYEAR': <bound method Func.from_arg_list of <class 'sqlglot.expressions.WeekOfYear'>>, 'WHEN': <bound method Func.from_arg_list of <class 'sqlglot.expressions.When'>>, 'X_M_L_TABLE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.XMLTable'>>, 'XOR': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Xor'>>, 'YEAR': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Year'>>, 'GLOB': <function Parser.<lambda>>, 'JSON_EXTRACT_PATH_TEXT': <function build_extract_json_with_path.<locals>._builder>, 'LIKE': <function build_like>, 'LOG2': <function Parser.<lambda>>, 'LOG10': <function Parser.<lambda>>, 'MOD': <function build_mod>, 'TO_HEX': <function build_hex>, 'ARBITRARY': <bound method Func.from_arg_list of <class 'sqlglot.expressions.AnyValue'>>, 'APPROX_PERCENTILE': <function _build_approx_percentile>, 'BITWISE_AND': <function binary_from_function.<locals>.<lambda>>, 'BITWISE_NOT': <function Presto.Parser.<lambda>>, 'BITWISE_OR': <function binary_from_function.<locals>.<lambda>>, 'BITWISE_XOR': <function binary_from_function.<locals>.<lambda>>, 'CARDINALITY': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArraySize'>>, 'CONTAINS': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArrayContains'>>, 'DATE_FORMAT': <function build_formatted_time.<locals>._builder>, 'DATE_PARSE': <function build_formatted_time.<locals>._builder>, 'ELEMENT_AT': <function Presto.Parser.<lambda>>, 'FROM_HEX': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Unhex'>>, 'FROM_UNIXTIME': <function _build_from_unixtime>, 'FROM_UTF8': <function Presto.Parser.<lambda>>, 'NOW': <bound method Func.from_arg_list of <class 'sqlglot.expressions.CurrentTimestamp'>>, 'ROW': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Struct'>>, 'SEQUENCE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.GenerateSeries'>>, 'SET_AGG': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArrayUniqueAgg'>>, 'SPLIT_TO_MAP': <bound method Func.from_arg_list of <class 'sqlglot.expressions.StrToMap'>>, 'STRPOS': <function Presto.Parser.<lambda>>, 'TO_UNIXTIME': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TimeToUnix'>>, 'TO_UTF8': <function Presto.Parser.<lambda>>, 'SHA256': <function Presto.Parser.<lambda>>, 'SHA512': <function Presto.Parser.<lambda>>}
FUNCTION_PARSERS = {'CAST': <function Parser.<lambda>>, 'CONVERT': <function Parser.<lambda>>, 'DECODE': <function Parser.<lambda>>, 'EXTRACT': <function Parser.<lambda>>, 'JSON_OBJECT': <function Parser.<lambda>>, 'JSON_OBJECTAGG': <function Parser.<lambda>>, 'JSON_TABLE': <function Parser.<lambda>>, 'MATCH': <function Parser.<lambda>>, 'OPENJSON': <function Parser.<lambda>>, 'POSITION': <function Parser.<lambda>>, 'PREDICT': <function Parser.<lambda>>, 'SAFE_CAST': <function Parser.<lambda>>, 'STRING_AGG': <function Parser.<lambda>>, 'SUBSTRING': <function Parser.<lambda>>, 'TRY_CAST': <function Parser.<lambda>>, 'TRY_CONVERT': <function Parser.<lambda>>}
TABLE_ALIAS_TOKENS = {<TokenType.KILL: 'KILL'>, <TokenType.MONEY: 'MONEY'>, <TokenType.COPY: 'COPY'>, <TokenType.PERCENT: 'PERCENT'>, <TokenType.TAG: 'TAG'>, <TokenType.TSTZMULTIRANGE: 'TSTZMULTIRANGE'>, <TokenType.VOLATILE: 'VOLATILE'>, <TokenType.IDENTIFIER: 'IDENTIFIER'>, <TokenType.VARBINARY: 'VARBINARY'>, <TokenType.HSTORE: 'HSTORE'>, <TokenType.ASC: 'ASC'>, <TokenType.IPADDRESS: 'IPADDRESS'>, <TokenType.ISNULL: 'ISNULL'>, <TokenType.TEMPORARY: 'TEMPORARY'>, <TokenType.PROCEDURE: 'PROCEDURE'>, <TokenType.SCHEMA: 'SCHEMA'>, <TokenType.NUMRANGE: 'NUMRANGE'>, <TokenType.DATE32: 'DATE32'>, <TokenType.INT8RANGE: 'INT8RANGE'>, <TokenType.ANTI: 'ANTI'>, <TokenType.AGGREGATEFUNCTION: 'AGGREGATEFUNCTION'>, <TokenType.VIEW: 'VIEW'>, <TokenType.VARCHAR: 'VARCHAR'>, <TokenType.TIMETZ: 'TIMETZ'>, <TokenType.GEOMETRY: 'GEOMETRY'>, <TokenType.BEGIN: 'BEGIN'>, <TokenType.ENUM8: 'ENUM8'>, <TokenType.INDEX: 'INDEX'>, <TokenType.PARTITION: 'PARTITION'>, <TokenType.UINT256: 'UINT256'>, <TokenType.BIGINT: 'BIGINT'>, <TokenType.CACHE: 'CACHE'>, <TokenType.OPERATOR: 'OPERATOR'>, <TokenType.COLUMN: 'COLUMN'>, <TokenType.INT128: 'INT128'>, <TokenType.SMALLMONEY: 'SMALLMONEY'>, <TokenType.HLLSKETCH: 'HLLSKETCH'>, <TokenType.UNPIVOT: 'UNPIVOT'>, <TokenType.SMALLINT: 'SMALLINT'>, <TokenType.INT256: 'INT256'>, <TokenType.FLOAT: 'FLOAT'>, <TokenType.COMMIT: 'COMMIT'>, <TokenType.NUMMULTIRANGE: 'NUMMULTIRANGE'>, <TokenType.EXISTS: 'EXISTS'>, <TokenType.NULL: 'NULL'>, <TokenType.RECURSIVE: 'RECURSIVE'>, <TokenType.SET: 'SET'>, <TokenType.NULLABLE: 'NULLABLE'>, <TokenType.REFRESH: 'REFRESH'>, <TokenType.TINYTEXT: 'TINYTEXT'>, <TokenType.SHOW: 'SHOW'>, <TokenType.DEFAULT: 'DEFAULT'>, <TokenType.DATE: 'DATE'>, <TokenType.TIMESTAMPLTZ: 'TIMESTAMPLTZ'>, <TokenType.MERGE: 'MERGE'>, <TokenType.FIRST: 'FIRST'>, <TokenType.BINARY: 'BINARY'>, <TokenType.ARRAY: 'ARRAY'>, <TokenType.SERIAL: 'SERIAL'>, <TokenType.EXECUTE: 'EXECUTE'>, <TokenType.TIMESTAMPNTZ: 'TIMESTAMPNTZ'>, <TokenType.INTERVAL: 'INTERVAL'>, <TokenType.PRAGMA: 'PRAGMA'>, <TokenType.STORAGE_INTEGRATION: 'STORAGE_INTEGRATION'>, <TokenType.BPCHAR: 'BPCHAR'>, <TokenType.BIT: 'BIT'>, <TokenType.UINT: 'UINT'>, <TokenType.UUID: 'UUID'>, <TokenType.STRUCT: 'STRUCT'>, <TokenType.FALSE: 'FALSE'>, <TokenType.FUNCTION: 'FUNCTION'>, <TokenType.TOP: 'TOP'>, <TokenType.SUPER: 'SUPER'>, <TokenType.CASE: 'CASE'>, <TokenType.IMAGE: 'IMAGE'>, <TokenType.UBIGINT: 'UBIGINT'>, <TokenType.LONGBLOB: 'LONGBLOB'>, <TokenType.ROLLUP: 'ROLLUP'>, <TokenType.VAR: 'VAR'>, <TokenType.COMMAND: 'COMMAND'>, <TokenType.LOAD: 'LOAD'>, <TokenType.TRUNCATE: 'TRUNCATE'>, <TokenType.KEEP: 'KEEP'>, <TokenType.UMEDIUMINT: 'UMEDIUMINT'>, <TokenType.OBJECT_IDENTIFIER: 'OBJECT_IDENTIFIER'>, <TokenType.UINT128: 'UINT128'>, <TokenType.END: 'END'>, <TokenType.MAP: 'MAP'>, <TokenType.IPV4: 'IPV4'>, <TokenType.USERDEFINED: 'USERDEFINED'>, <TokenType.REFERENCES: 'REFERENCES'>, <TokenType.REPLACE: 'REPLACE'>, <TokenType.RANGE: 'RANGE'>, <TokenType.NVARCHAR: 'NVARCHAR'>, <TokenType.INT4MULTIRANGE: 'INT4MULTIRANGE'>, <TokenType.TRUE: 'TRUE'>, <TokenType.TIMESTAMP_MS: 'TIMESTAMP_MS'>, <TokenType.VARIANT: 'VARIANT'>, <TokenType.ENUM: 'ENUM'>, <TokenType.USE: 'USE'>, <TokenType.USMALLINT: 'USMALLINT'>, <TokenType.OVERLAPS: 'OVERLAPS'>, <TokenType.TIMESTAMP_S: 'TIMESTAMP_S'>, <TokenType.CURRENT_TIMESTAMP: 'CURRENT_TIMESTAMP'>, <TokenType.UNIQUE: 'UNIQUE'>, <TokenType.ORDINALITY: 'ORDINALITY'>, <TokenType.UNIQUEIDENTIFIER: 'UNIQUEIDENTIFIER'>, <TokenType.TEXT: 'TEXT'>, <TokenType.DATABASE: 'DATABASE'>, <TokenType.BIGDECIMAL: 'BIGDECIMAL'>, <TokenType.PSEUDO_TYPE: 'PSEUDO_TYPE'>, <TokenType.FIXEDSTRING: 'FIXEDSTRING'>, <TokenType.NEXT: 'NEXT'>, <TokenType.INT: 'INT'>, <TokenType.BIGSERIAL: 'BIGSERIAL'>, <TokenType.BOOLEAN: 'BOOLEAN'>, <TokenType.IS: 'IS'>, <TokenType.CONSTRAINT: 'CONSTRAINT'>, <TokenType.FINAL: 'FINAL'>, <TokenType.UNKNOWN: 'UNKNOWN'>, <TokenType.SETTINGS: 'SETTINGS'>, <TokenType.ROWS: 'ROWS'>, <TokenType.TSTZRANGE: 'TSTZRANGE'>, <TokenType.DATETIME64: 'DATETIME64'>, <TokenType.MEDIUMBLOB: 'MEDIUMBLOB'>, <TokenType.DATETIME: 'DATETIME'>, <TokenType.MEDIUMINT: 'MEDIUMINT'>, <TokenType.TIMESTAMP: 'TIMESTAMP'>, <TokenType.TSMULTIRANGE: 'TSMULTIRANGE'>, <TokenType.XML: 'XML'>, <TokenType.JSON: 'JSON'>, <TokenType.UTINYINT: 'UTINYINT'>, <TokenType.DELETE: 'DELETE'>, <TokenType.TIME: 'TIME'>, <TokenType.UDECIMAL: 'UDECIMAL'>, <TokenType.NESTED: 'NESTED'>, <TokenType.SEMI: 'SEMI'>, <TokenType.DESC: 'DESC'>, <TokenType.ROWVERSION: 'ROWVERSION'>, <TokenType.TIMESTAMP_NS: 'TIMESTAMP_NS'>, <TokenType.ANY: 'ANY'>, <TokenType.DIV: 'DIV'>, <TokenType.DICTIONARY: 'DICTIONARY'>, <TokenType.INET: 'INET'>, <TokenType.INT8MULTIRANGE: 'INT8MULTIRANGE'>, <TokenType.COMMENT: 'COMMENT'>, <TokenType.CURRENT_USER: 'CURRENT_USER'>, <TokenType.SIMPLEAGGREGATEFUNCTION: 'SIMPLEAGGREGATEFUNCTION'>, <TokenType.ALL: 'ALL'>, <TokenType.NAME: 'NAME'>, <TokenType.SOME: 'SOME'>, <TokenType.ENUM16: 'ENUM16'>, <TokenType.YEAR: 'YEAR'>, <TokenType.GEOGRAPHY: 'GEOGRAPHY'>, <TokenType.DOUBLE: 'DOUBLE'>, <TokenType.INT4RANGE: 'INT4RANGE'>, <TokenType.FORMAT: 'FORMAT'>, <TokenType.FOREIGN_KEY: 'FOREIGN_KEY'>, <TokenType.AUTO_INCREMENT: 'AUTO_INCREMENT'>, <TokenType.IPPREFIX: 'IPPREFIX'>, <TokenType.OVERWRITE: 'OVERWRITE'>, <TokenType.CURRENT_DATETIME: 'CURRENT_DATETIME'>, <TokenType.COLLATE: 'COLLATE'>, <TokenType.FILTER: 'FILTER'>, <TokenType.JSONB: 'JSONB'>, <TokenType.TDIGEST: 'TDIGEST'>, <TokenType.CHAR: 'CHAR'>, <TokenType.MEDIUMTEXT: 'MEDIUMTEXT'>, <TokenType.LOWCARDINALITY: 'LOWCARDINALITY'>, <TokenType.ROW: 'ROW'>, <TokenType.SMALLSERIAL: 'SMALLSERIAL'>, <TokenType.OBJECT: 'OBJECT'>, <TokenType.TIMESTAMPTZ: 'TIMESTAMPTZ'>, <TokenType.TABLE: 'TABLE'>, <TokenType.DATERANGE: 'DATERANGE'>, <TokenType.TSRANGE: 'TSRANGE'>, <TokenType.ESCAPE: 'ESCAPE'>, <TokenType.PIVOT: 'PIVOT'>, <TokenType.DECIMAL: 'DECIMAL'>, <TokenType.NCHAR: 'NCHAR'>, <TokenType.DATEMULTIRANGE: 'DATEMULTIRANGE'>, <TokenType.UPDATE: 'UPDATE'>, <TokenType.MODEL: 'MODEL'>, <TokenType.CURRENT_DATE: 'CURRENT_DATE'>, <TokenType.SEQUENCE: 'SEQUENCE'>, <TokenType.LONGTEXT: 'LONGTEXT'>, <TokenType.DESCRIBE: 'DESCRIBE'>, <TokenType.IPV6: 'IPV6'>, <TokenType.CURRENT_TIME: 'CURRENT_TIME'>, <TokenType.TINYINT: 'TINYINT'>, <TokenType.TINYBLOB: 'TINYBLOB'>}
SHOW_TRIE: Dict = {}
SET_TRIE: Dict = {'GLOBAL': {0: True}, 'LOCAL': {0: True}, 'SESSION': {0: True}, 'TRANSACTION': {0: True}}

##### Inherited Members

sqlglot.parser.Parser
  Parser
  NO_PAREN_FUNCTIONS
  STRUCT_TYPE_TOKENS
  NESTED_TYPE_TOKENS
  ENUM_TYPE_TOKENS
  AGGREGATE_TYPE_TOKENS
  TYPE_TOKENS
  SIGNED_TO_UNSIGNED_TYPE_TOKEN
  SUBQUERY_PREDICATES
  RESERVED_TOKENS
  DB_CREATABLES
  CREATABLES
  ID_VAR_TOKENS
  INTERVAL_VARS
  ALIAS_TOKENS
  COMMENT_TABLE_ALIAS_TOKENS
  UPDATE_ALIAS_TOKENS
  TRIM_TYPES
  FUNC_TOKENS
  CONJUNCTION
  EQUALITY
  COMPARISON
  BITWISE
  TERM
  FACTOR
  EXPONENT
  TIMES
  TIMESTAMPS
  SET_OPERATIONS
  JOIN_METHODS
  JOIN_SIDES
  JOIN_KINDS
  JOIN_HINTS
  LAMBDAS
  COLUMN_OPERATORS
  EXPRESSION_PARSERS
  STATEMENT_PARSERS
  UNARY_PARSERS
  STRING_PARSERS
  NUMERIC_PARSERS
  PRIMARY_PARSERS
  PLACEHOLDER_PARSERS
  RANGE_PARSERS
  PROPERTY_PARSERS
  CONSTRAINT_PARSERS
  ALTER_PARSERS
  ALTER_ALTER_PARSERS
  SCHEMA_UNNAMED_CONSTRAINTS
  NO_PAREN_FUNCTION_PARSERS
  INVALID_FUNC_NAME_TOKENS
  FUNCTIONS_WITH_ALIASED_ARGS
  KEY_VALUE_DEFINITIONS
  QUERY_MODIFIER_PARSERS
  SET_PARSERS
  SHOW_PARSERS
  TYPE_LITERAL_PARSERS
  TYPE_CONVERTER
  DDL_SELECT_TOKENS
  PRE_VOLATILE_TOKENS
  TRANSACTION_KIND
  TRANSACTION_CHARACTERISTICS
  CONFLICT_ACTIONS
  CREATE_SEQUENCE
  ISOLATED_LOADING_OPTIONS
  USABLES
  CAST_ACTIONS
  INSERT_ALTERNATIVES
  CLONE_KEYWORDS
  HISTORICAL_DATA_KIND
  OPCLASS_FOLLOW_KEYWORDS
  OPTYPE_FOLLOW_TOKENS
  TABLE_INDEX_HINT_TOKENS
  VIEW_ATTRIBUTES
  WINDOW_ALIAS_TOKENS
  WINDOW_BEFORE_PAREN_TOKENS
  WINDOW_SIDES
  JSON_KEY_VALUE_SEPARATOR_TOKENS
  FETCH_TOKENS
  ADD_CONSTRAINT_TOKENS
  DISTINCT_TOKENS
  NULL_TOKENS
  UNNEST_OFFSET_ALIAS_TOKENS
  SELECT_START_TOKENS
  STRICT_CAST
  PREFIXED_PIVOT_COLUMNS
  IDENTIFY_PIVOT_STRINGS
  LOG_DEFAULTS_TO_LN
  ALTER_TABLE_ADD_REQUIRED_FOR_EACH_COLUMN
  TABLESAMPLE_CSV
  DEFAULT_SAMPLING_METHOD
  SET_REQUIRES_ASSIGNMENT_DELIMITER
  TRIM_PATTERN_FIRST
  STRING_ALIASES
  MODIFIERS_ATTACHED_TO_UNION
  UNION_MODIFIERS
  NO_PAREN_IF_COMMANDS
  JSON_ARROWS_REQUIRE_JSON_TYPE
  COLON_IS_JSON_EXTRACT
  SUPPORTS_IMPLICIT_UNNEST
  INTERVAL_SPANS
  SUPPORTS_PARTITION_SELECTION
  error_level
  error_message_context
  max_errors
  dialect
  reset
  parse
  parse_into
  check_errors
  raise_error
  expression
  validate_expression
  errors
  sql

class Presto.Generator(sqlglot.generator.Generator): View Source

    293    class Generator(generator.Generator):
    294        INTERVAL_ALLOWS_PLURAL_FORM = False
    295        JOIN_HINTS = False
    296        TABLE_HINTS = False
    297        QUERY_HINTS = False
    298        IS_BOOL_ALLOWED = False
    299        TZ_TO_WITH_TIME_ZONE = True
    300        NVL2_SUPPORTED = False
    301        STRUCT_DELIMITER = ("(", ")")
    302        LIMIT_ONLY_LITERALS = True
    303        SUPPORTS_SINGLE_ARG_CONCAT = False
    304        LIKE_PROPERTY_INSIDE_SCHEMA = True
    305        MULTI_ARG_DISTINCT = False
    306        SUPPORTS_TO_NUMBER = False
    307        HEX_FUNC = "TO_HEX"
    308
    309        PROPERTIES_LOCATION = {
    310            **generator.Generator.PROPERTIES_LOCATION,
    311            exp.LocationProperty: exp.Properties.Location.UNSUPPORTED,
    312            exp.VolatileProperty: exp.Properties.Location.UNSUPPORTED,
    313        }
    314
    315        TYPE_MAPPING = {
    316            **generator.Generator.TYPE_MAPPING,
    317            exp.DataType.Type.INT: "INTEGER",
    318            exp.DataType.Type.FLOAT: "REAL",
    319            exp.DataType.Type.BINARY: "VARBINARY",
    320            exp.DataType.Type.TEXT: "VARCHAR",
    321            exp.DataType.Type.TIMETZ: "TIME",
    322            exp.DataType.Type.TIMESTAMPTZ: "TIMESTAMP",
    323            exp.DataType.Type.STRUCT: "ROW",
    324            exp.DataType.Type.DATETIME: "TIMESTAMP",
    325            exp.DataType.Type.DATETIME64: "TIMESTAMP",
    326            exp.DataType.Type.HLLSKETCH: "HYPERLOGLOG",
    327        }
    328
    329        TRANSFORMS = {
    330            **generator.Generator.TRANSFORMS,
    331            exp.AnyValue: rename_func("ARBITRARY"),
    332            exp.ApproxDistinct: lambda self, e: self.func(
    333                "APPROX_DISTINCT", e.this, e.args.get("accuracy")
    334            ),
    335            exp.ApproxQuantile: rename_func("APPROX_PERCENTILE"),
    336            exp.ArgMax: rename_func("MAX_BY"),
    337            exp.ArgMin: rename_func("MIN_BY"),
    338            exp.Array: lambda self, e: f"ARRAY[{self.expressions(e, flat=True)}]",
    339            exp.ArrayAny: rename_func("ANY_MATCH"),
    340            exp.ArrayConcat: rename_func("CONCAT"),
    341            exp.ArrayContains: rename_func("CONTAINS"),
    342            exp.ArraySize: rename_func("CARDINALITY"),
    343            exp.ArrayToString: rename_func("ARRAY_JOIN"),
    344            exp.ArrayUniqueAgg: rename_func("SET_AGG"),
    345            exp.AtTimeZone: rename_func("AT_TIMEZONE"),
    346            exp.BitwiseAnd: lambda self, e: self.func("BITWISE_AND", e.this, e.expression),
    347            exp.BitwiseLeftShift: lambda self, e: self.func(
    348                "BITWISE_ARITHMETIC_SHIFT_LEFT", e.this, e.expression
    349            ),
    350            exp.BitwiseNot: lambda self, e: self.func("BITWISE_NOT", e.this),
    351            exp.BitwiseOr: lambda self, e: self.func("BITWISE_OR", e.this, e.expression),
    352            exp.BitwiseRightShift: lambda self, e: self.func(
    353                "BITWISE_ARITHMETIC_SHIFT_RIGHT", e.this, e.expression
    354            ),
    355            exp.BitwiseXor: lambda self, e: self.func("BITWISE_XOR", e.this, e.expression),
    356            exp.Cast: transforms.preprocess([transforms.epoch_cast_to_ts]),
    357            exp.CurrentTimestamp: lambda *_: "CURRENT_TIMESTAMP",
    358            exp.DateAdd: lambda self, e: self.func(
    359                "DATE_ADD",
    360                unit_to_str(e),
    361                _to_int(e.expression),
    362                e.this,
    363            ),
    364            exp.DateDiff: lambda self, e: self.func(
    365                "DATE_DIFF", unit_to_str(e), e.expression, e.this
    366            ),
    367            exp.DateStrToDate: datestrtodate_sql,
    368            exp.DateToDi: lambda self,
    369            e: f"CAST(DATE_FORMAT({self.sql(e, 'this')}, {Presto.DATEINT_FORMAT}) AS INT)",
    370            exp.DateSub: lambda self, e: self.func(
    371                "DATE_ADD",
    372                unit_to_str(e),
    373                _to_int(e.expression * -1),
    374                e.this,
    375            ),
    376            exp.Decode: lambda self, e: encode_decode_sql(self, e, "FROM_UTF8"),
    377            exp.DiToDate: lambda self,
    378            e: f"CAST(DATE_PARSE(CAST({self.sql(e, 'this')} AS VARCHAR), {Presto.DATEINT_FORMAT}) AS DATE)",
    379            exp.Encode: lambda self, e: encode_decode_sql(self, e, "TO_UTF8"),
    380            exp.FileFormatProperty: lambda self, e: f"FORMAT='{e.name.upper()}'",
    381            exp.First: _first_last_sql,
    382            exp.FirstValue: _first_last_sql,
    383            exp.FromTimeZone: lambda self,
    384            e: f"WITH_TIMEZONE({self.sql(e, 'this')}, {self.sql(e, 'zone')}) AT TIME ZONE 'UTC'",
    385            exp.Group: transforms.preprocess([transforms.unalias_group]),
    386            exp.GroupConcat: lambda self, e: self.func(
    387                "ARRAY_JOIN", self.func("ARRAY_AGG", e.this), e.args.get("separator")
    388            ),
    389            exp.If: if_sql(),
    390            exp.ILike: no_ilike_sql,
    391            exp.Initcap: _initcap_sql,
    392            exp.ParseJSON: rename_func("JSON_PARSE"),
    393            exp.Last: _first_last_sql,
    394            exp.LastValue: _first_last_sql,
    395            exp.LastDay: lambda self, e: self.func("LAST_DAY_OF_MONTH", e.this),
    396            exp.Lateral: _explode_to_unnest_sql,
    397            exp.Left: left_to_substring_sql,
    398            exp.Levenshtein: rename_func("LEVENSHTEIN_DISTANCE"),
    399            exp.LogicalAnd: rename_func("BOOL_AND"),
    400            exp.LogicalOr: rename_func("BOOL_OR"),
    401            exp.Pivot: no_pivot_sql,
    402            exp.Quantile: _quantile_sql,
    403            exp.RegexpExtract: regexp_extract_sql,
    404            exp.Right: right_to_substring_sql,
    405            exp.SafeDivide: no_safe_divide_sql,
    406            exp.Schema: _schema_sql,
    407            exp.SchemaCommentProperty: lambda self, e: self.naked_property(e),
    408            exp.Select: transforms.preprocess(
    409                [
    410                    transforms.eliminate_qualify,
    411                    transforms.eliminate_distinct_on,
    412                    transforms.explode_to_unnest(1),
    413                    transforms.eliminate_semi_and_anti_joins,
    414                ]
    415            ),
    416            exp.SortArray: _no_sort_array,
    417            exp.StrPosition: lambda self, e: str_position_sql(self, e, generate_instance=True),
    418            exp.StrToDate: lambda self, e: f"CAST({_str_to_time_sql(self, e)} AS DATE)",
    419            exp.StrToMap: rename_func("SPLIT_TO_MAP"),
    420            exp.StrToTime: _str_to_time_sql,
    421            exp.StructExtract: struct_extract_sql,
    422            exp.Table: transforms.preprocess([_unnest_sequence]),
    423            exp.Timestamp: no_timestamp_sql,
    424            exp.TimestampTrunc: timestamptrunc_sql(),
    425            exp.TimeStrToDate: timestrtotime_sql,
    426            exp.TimeStrToTime: timestrtotime_sql,
    427            exp.TimeStrToUnix: lambda self, e: self.func(
    428                "TO_UNIXTIME", self.func("DATE_PARSE", e.this, Presto.TIME_FORMAT)
    429            ),
    430            exp.TimeToStr: lambda self, e: self.func("DATE_FORMAT", e.this, self.format_time(e)),
    431            exp.TimeToUnix: rename_func("TO_UNIXTIME"),
    432            exp.ToChar: lambda self, e: self.func("DATE_FORMAT", e.this, self.format_time(e)),
    433            exp.TryCast: transforms.preprocess([transforms.epoch_cast_to_ts]),
    434            exp.TsOrDiToDi: lambda self,
    435            e: f"CAST(SUBSTR(REPLACE(CAST({self.sql(e, 'this')} AS VARCHAR), '-', ''), 1, 8) AS INT)",
    436            exp.TsOrDsAdd: _ts_or_ds_add_sql,
    437            exp.TsOrDsDiff: _ts_or_ds_diff_sql,
    438            exp.TsOrDsToDate: _ts_or_ds_to_date_sql,
    439            exp.Unhex: rename_func("FROM_HEX"),
    440            exp.UnixToStr: lambda self,
    441            e: f"DATE_FORMAT(FROM_UNIXTIME({self.sql(e, 'this')}), {self.format_time(e)})",
    442            exp.UnixToTime: _unix_to_time_sql,
    443            exp.UnixToTimeStr: lambda self,
    444            e: f"CAST(FROM_UNIXTIME({self.sql(e, 'this')}) AS VARCHAR)",
    445            exp.VariancePop: rename_func("VAR_POP"),
    446            exp.With: transforms.preprocess([transforms.add_recursive_cte_column_names]),
    447            exp.WithinGroup: transforms.preprocess(
    448                [transforms.remove_within_group_for_percentiles]
    449            ),
    450            exp.Xor: bool_xor_sql,
    451            exp.MD5: lambda self, e: self.func(
    452                "LOWER", self.func("TO_HEX", self.func("MD5", self.sql(e, "this")))
    453            ),
    454            exp.MD5Digest: rename_func("MD5"),
    455            exp.SHA: rename_func("SHA1"),
    456            exp.SHA2: lambda self, e: self.func(
    457                "SHA256" if e.text("length") == "256" else "SHA512", e.this
    458            ),
    459        }
    460
    461        RESERVED_KEYWORDS = {
    462            "alter",
    463            "and",
    464            "as",
    465            "between",
    466            "by",
    467            "case",
    468            "cast",
    469            "constraint",
    470            "create",
    471            "cross",
    472            "current_time",
    473            "current_timestamp",
    474            "deallocate",
    475            "delete",
    476            "describe",
    477            "distinct",
    478            "drop",
    479            "else",
    480            "end",
    481            "escape",
    482            "except",
    483            "execute",
    484            "exists",
    485            "extract",
    486            "false",
    487            "for",
    488            "from",
    489            "full",
    490            "group",
    491            "having",
    492            "in",
    493            "inner",
    494            "insert",
    495            "intersect",
    496            "into",
    497            "is",
    498            "join",
    499            "left",
    500            "like",
    501            "natural",
    502            "not",
    503            "null",
    504            "on",
    505            "or",
    506            "order",
    507            "outer",
    508            "prepare",
    509            "right",
    510            "select",
    511            "table",
    512            "then",
    513            "true",
    514            "union",
    515            "using",
    516            "values",
    517            "when",
    518            "where",
    519            "with",
    520        }
    521
    522        def strtounix_sql(self, expression: exp.StrToUnix) -> str:
    523            # Since `TO_UNIXTIME` requires a `TIMESTAMP`, we need to parse the argument into one.
    524            # To do this, we first try to `DATE_PARSE` it, but since this can fail when there's a
    525            # timezone involved, we wrap it in a `TRY` call and use `PARSE_DATETIME` as a fallback,
    526            # which seems to be using the same time mapping as Hive, as per:
    527            # https://joda-time.sourceforge.net/apidocs/org/joda/time/format/DateTimeFormat.html
    528            value_as_text = exp.cast(expression.this, exp.DataType.Type.TEXT)
    529            parse_without_tz = self.func("DATE_PARSE", value_as_text, self.format_time(expression))
    530            parse_with_tz = self.func(
    531                "PARSE_DATETIME",
    532                value_as_text,
    533                self.format_time(expression, Hive.INVERSE_TIME_MAPPING, Hive.INVERSE_TIME_TRIE),
    534            )
    535            coalesced = self.func("COALESCE", self.func("TRY", parse_without_tz), parse_with_tz)
    536            return self.func("TO_UNIXTIME", coalesced)
    537
    538        def bracket_sql(self, expression: exp.Bracket) -> str:
    539            if expression.args.get("safe"):
    540                return self.func(
    541                    "ELEMENT_AT",
    542                    expression.this,
    543                    seq_get(
    544                        apply_index_offset(
    545                            expression.this,
    546                            expression.expressions,
    547                            1 - expression.args.get("offset", 0),
    548                        ),
    549                        0,
    550                    ),
    551                )
    552            return super().bracket_sql(expression)
    553
    554        def struct_sql(self, expression: exp.Struct) -> str:
    555            from sqlglot.optimizer.annotate_types import annotate_types
    556
    557            expression = annotate_types(expression)
    558            values: t.List[str] = []
    559            schema: t.List[str] = []
    560            unknown_type = False
    561
    562            for e in expression.expressions:
    563                if isinstance(e, exp.PropertyEQ):
    564                    if e.type and e.type.is_type(exp.DataType.Type.UNKNOWN):
    565                        unknown_type = True
    566                    else:
    567                        schema.append(f"{self.sql(e, 'this')} {self.sql(e.type)}")
    568                    values.append(self.sql(e, "expression"))
    569                else:
    570                    values.append(self.sql(e))
    571
    572            size = len(expression.expressions)
    573
    574            if not size or len(schema) != size:
    575                if unknown_type:
    576                    self.unsupported(
    577                        "Cannot convert untyped key-value definitions (try annotate_types)."
    578                    )
    579                return self.func("ROW", *values)
    580            return f"CAST(ROW({', '.join(values)}) AS ROW({', '.join(schema)}))"
    581
    582        def interval_sql(self, expression: exp.Interval) -> str:
    583            if expression.this and expression.text("unit").upper().startswith("WEEK"):
    584                return f"({expression.this.name} * INTERVAL '7' DAY)"
    585            return super().interval_sql(expression)
    586
    587        def transaction_sql(self, expression: exp.Transaction) -> str:
    588            modes = expression.args.get("modes")
    589            modes = f" {', '.join(modes)}" if modes else ""
    590            return f"START TRANSACTION{modes}"
    591
    592        def generateseries_sql(self, expression: exp.GenerateSeries) -> str:
    593            start = expression.args["start"]
    594            end = expression.args["end"]
    595            step = expression.args.get("step")
    596
    597            if isinstance(start, exp.Cast):
    598                target_type = start.to
    599            elif isinstance(end, exp.Cast):
    600                target_type = end.to
    601            else:
    602                target_type = None
    603
    604            if target_type and target_type.is_type("timestamp"):
    605                if target_type is start.to:
    606                    end = exp.cast(end, target_type)
    607                else:
    608                    start = exp.cast(start, target_type)
    609
    610            return self.func("SEQUENCE", start, end, step)
    611
    612        def offset_limit_modifiers(
    613            self, expression: exp.Expression, fetch: bool, limit: t.Optional[exp.Fetch | exp.Limit]
    614        ) -> t.List[str]:
    615            return [
    616                self.sql(expression, "offset"),
    617                self.sql(limit),
    618            ]
    619
    620        def create_sql(self, expression: exp.Create) -> str:
    621            """
    622            Presto doesn't support CREATE VIEW with expressions (ex: `CREATE VIEW x (cola)` then `(cola)` is the expression),
    623            so we need to remove them
    624            """
    625            kind = expression.args["kind"]
    626            schema = expression.this
    627            if kind == "VIEW" and schema.expressions:
    628                expression.this.set("expressions", None)
    629            return super().create_sql(expression)
    630
    631        def delete_sql(self, expression: exp.Delete) -> str:
    632            """
    633            Presto only supports DELETE FROM for a single table without an alias, so we need
    634            to remove the unnecessary parts. If the original DELETE statement contains more
    635            than one table to be deleted, we can't safely map it 1-1 to a Presto statement.
    636            """
    637            tables = expression.args.get("tables") or [expression.this]
    638            if len(tables) > 1:
    639                return super().delete_sql(expression)
    640
    641            table = tables[0]
    642            expression.set("this", table)
    643            expression.set("tables", None)
    644
    645            if isinstance(table, exp.Table):
    646                table_alias = table.args.get("alias")
    647                if table_alias:
    648                    table_alias.pop()
    649                    expression = t.cast(exp.Delete, expression.transform(unqualify_columns))
    650
    651            return super().delete_sql(expression)

Generator converts a given syntax tree to the corresponding SQL string.

###### Arguments:

- pretty: Whether to format the produced SQL string. Default: False.
- identify: Determines when an identifier should be quoted. Possible values are: False (default): Never quote, except in cases where it's mandatory by the dialect. True or 'always': Always quote. 'safe': Only quote identifiers that are case insensitive.
- normalize: Whether to normalize identifiers to lowercase. Default: False.
- pad: The pad size in a formatted string. For example, this affects the indentation of a projection in a query, relative to its nesting level. Default: 2.
- indent: The indentation size in a formatted string. For example, this affects the indentation of subqueries and filters under a `WHERE` clause. Default: 2.
- normalize_functions: How to normalize function names. Possible values are: "upper" or True (default): Convert names to uppercase. "lower": Convert names to lowercase. False: Disables function name normalization.
- unsupported_level: Determines the generator's behavior when it encounters unsupported expressions. Default ErrorLevel.WARN.
- max_unsupported: Maximum number of unsupported messages to include in a raised UnsupportedError. This is only relevant if unsupported_level is ErrorLevel.RAISE. Default: 3
- leading_comma: Whether the comma is leading or trailing in select expressions. This is only relevant when generating in pretty mode. Default: False
- max_text_width: The max number of characters in a segment before creating new lines in pretty mode. The default is on the smaller end because the length only represents a segment and not the true line length. Default: 80
- comments: Whether to preserve comments in the output SQL code. Default: True

INTERVAL_ALLOWS_PLURAL_FORM = False
JOIN_HINTS = False
TABLE_HINTS = False
QUERY_HINTS = False
IS_BOOL_ALLOWED = False
TZ_TO_WITH_TIME_ZONE = True
NVL2_SUPPORTED = False
STRUCT_DELIMITER = ('(', ')')
LIMIT_ONLY_LITERALS = True
SUPPORTS_SINGLE_ARG_CONCAT = False
LIKE_PROPERTY_INSIDE_SCHEMA = True
MULTI_ARG_DISTINCT = False
SUPPORTS_TO_NUMBER = False
HEX_FUNC = 'TO_HEX'
PROPERTIES_LOCATION = {<class 'sqlglot.expressions.AllowedValuesProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.AlgorithmProperty'>: <Location.POST_CREATE: 'POST_CREATE'>, <class 'sqlglot.expressions.AutoIncrementProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.AutoRefreshProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.BackupProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.BlockCompressionProperty'>: <Location.POST_NAME: 'POST_NAME'>, <class 'sqlglot.expressions.CharacterSetProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.ChecksumProperty'>: <Location.POST_NAME: 'POST_NAME'>, <class 'sqlglot.expressions.CollateProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.CopyGrantsProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.Cluster'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.ClusteredByProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.DataBlocksizeProperty'>: <Location.POST_NAME: 'POST_NAME'>, <class 'sqlglot.expressions.DataDeletionProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.DefinerProperty'>: <Location.POST_CREATE: 'POST_CREATE'>, <class 'sqlglot.expressions.DictRange'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.DictProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.DistKeyProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.DistStyleProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.EngineProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.ExecuteAsProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.ExternalProperty'>: <Location.POST_CREATE: 'POST_CREATE'>, <class 'sqlglot.expressions.FallbackProperty'>: <Location.POST_NAME: 'POST_NAME'>, <class 'sqlglot.expressions.FileFormatProperty'>: <Location.POST_WITH: 'POST_WITH'>, <class 'sqlglot.expressions.FreespaceProperty'>: <Location.POST_NAME: 'POST_NAME'>, <class 'sqlglot.expressions.GlobalProperty'>: <Location.POST_CREATE: 'POST_CREATE'>, <class 'sqlglot.expressions.HeapProperty'>: <Location.POST_WITH: 'POST_WITH'>, <class 'sqlglot.expressions.InheritsProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.IcebergProperty'>: <Location.POST_CREATE: 'POST_CREATE'>, <class 'sqlglot.expressions.InputModelProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.IsolatedLoadingProperty'>: <Location.POST_NAME: 'POST_NAME'>, <class 'sqlglot.expressions.JournalProperty'>: <Location.POST_NAME: 'POST_NAME'>, <class 'sqlglot.expressions.LanguageProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.LikeProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.LocationProperty'>: <Location.UNSUPPORTED: 'UNSUPPORTED'>, <class 'sqlglot.expressions.LockProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.LockingProperty'>: <Location.POST_ALIAS: 'POST_ALIAS'>, <class 'sqlglot.expressions.LogProperty'>: <Location.POST_NAME: 'POST_NAME'>, <class 'sqlglot.expressions.MaterializedProperty'>: <Location.POST_CREATE: 'POST_CREATE'>, <class 'sqlglot.expressions.MergeBlockRatioProperty'>: <Location.POST_NAME: 'POST_NAME'>, <class 'sqlglot.expressions.NoPrimaryIndexProperty'>: <Location.POST_EXPRESSION: 'POST_EXPRESSION'>, <class 'sqlglot.expressions.OnProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.OnCommitProperty'>: <Location.POST_EXPRESSION: 'POST_EXPRESSION'>, <class 'sqlglot.expressions.Order'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.OutputModelProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.PartitionedByProperty'>: <Location.POST_WITH: 'POST_WITH'>, <class 'sqlglot.expressions.PartitionedOfProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.PrimaryKey'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.Property'>: <Location.POST_WITH: 'POST_WITH'>, <class 'sqlglot.expressions.RemoteWithConnectionModelProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.ReturnsProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.RowFormatProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.RowFormatDelimitedProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.RowFormatSerdeProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.SampleProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.SchemaCommentProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.SerdeProperties'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.Set'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.SettingsProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.SetProperty'>: <Location.POST_CREATE: 'POST_CREATE'>, <class 'sqlglot.expressions.SetConfigProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.SharingProperty'>: <Location.POST_EXPRESSION: 'POST_EXPRESSION'>, <class 'sqlglot.expressions.SequenceProperties'>: <Location.POST_EXPRESSION: 'POST_EXPRESSION'>, <class 'sqlglot.expressions.SortKeyProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.SqlReadWriteProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.SqlSecurityProperty'>: <Location.POST_CREATE: 'POST_CREATE'>, <class 'sqlglot.expressions.StabilityProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.StrictProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.TemporaryProperty'>: <Location.POST_CREATE: 'POST_CREATE'>, <class 'sqlglot.expressions.ToTableProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.TransientProperty'>: <Location.POST_CREATE: 'POST_CREATE'>, <class 'sqlglot.expressions.TransformModelProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.MergeTreeTTL'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.UnloggedProperty'>: <Location.POST_CREATE: 'POST_CREATE'>, <class 'sqlglot.expressions.ViewAttributeProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.VolatileProperty'>: <Location.UNSUPPORTED: 'UNSUPPORTED'>, <class 'sqlglot.expressions.WithDataProperty'>: <Location.POST_EXPRESSION: 'POST_EXPRESSION'>, <class 'sqlglot.expressions.WithJournalTableProperty'>: <Location.POST_NAME: 'POST_NAME'>, <class 'sqlglot.expressions.WithSystemVersioningProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>}
TYPE_MAPPING = {<Type.NCHAR: 'NCHAR'>: 'CHAR', <Type.NVARCHAR: 'NVARCHAR'>: 'VARCHAR', <Type.MEDIUMTEXT: 'MEDIUMTEXT'>: 'TEXT', <Type.LONGTEXT: 'LONGTEXT'>: 'TEXT', <Type.TINYTEXT: 'TINYTEXT'>: 'TEXT', <Type.MEDIUMBLOB: 'MEDIUMBLOB'>: 'BLOB', <Type.LONGBLOB: 'LONGBLOB'>: 'BLOB', <Type.TINYBLOB: 'TINYBLOB'>: 'BLOB', <Type.INET: 'INET'>: 'INET', <Type.ROWVERSION: 'ROWVERSION'>: 'VARBINARY', <Type.INT: 'INT'>: 'INTEGER', <Type.FLOAT: 'FLOAT'>: 'REAL', <Type.BINARY: 'BINARY'>: 'VARBINARY', <Type.TEXT: 'TEXT'>: 'VARCHAR', <Type.TIMETZ: 'TIMETZ'>: 'TIME', <Type.TIMESTAMPTZ: 'TIMESTAMPTZ'>: 'TIMESTAMP', <Type.STRUCT: 'STRUCT'>: 'ROW', <Type.DATETIME: 'DATETIME'>: 'TIMESTAMP', <Type.DATETIME64: 'DATETIME64'>: 'TIMESTAMP', <Type.HLLSKETCH: 'HLLSKETCH'>: 'HYPERLOGLOG'}
TRANSFORMS = {<class 'sqlglot.expressions.JSONPathFilter'>: <function <lambda>>, <class 'sqlglot.expressions.JSONPathKey'>: <function <lambda>>, <class 'sqlglot.expressions.JSONPathRecursive'>: <function <lambda>>, <class 'sqlglot.expressions.JSONPathRoot'>: <function <lambda>>, <class 'sqlglot.expressions.JSONPathScript'>: <function <lambda>>, <class 'sqlglot.expressions.JSONPathSelector'>: <function <lambda>>, <class 'sqlglot.expressions.JSONPathSlice'>: <function <lambda>>, <class 'sqlglot.expressions.JSONPathSubscript'>: <function <lambda>>, <class 'sqlglot.expressions.JSONPathUnion'>: <function <lambda>>, <class 'sqlglot.expressions.JSONPathWildcard'>: <function <lambda>>, <class 'sqlglot.expressions.AllowedValuesProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.AutoRefreshProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.BackupProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.CaseSpecificColumnConstraint'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.CharacterSetColumnConstraint'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.CharacterSetProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.ClusteredColumnConstraint'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.CollateColumnConstraint'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.CommentColumnConstraint'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.CopyGrantsProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.DateFormatColumnConstraint'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.DefaultColumnConstraint'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.EncodeColumnConstraint'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.EphemeralColumnConstraint'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.ExcludeColumnConstraint'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.ExecuteAsProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.ExternalProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.GlobalProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.HeapProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.IcebergProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.InheritsProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.InlineLengthColumnConstraint'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.InputModelProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.IntervalSpan'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.JSONExtract'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.JSONExtractScalar'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.LanguageProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.LocationProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.LogProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.MaterializedProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.NonClusteredColumnConstraint'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.NoPrimaryIndexProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.NotForReplicationColumnConstraint'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.OnCommitProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.OnProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.OnUpdateColumnConstraint'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.OutputModelProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.PathColumnConstraint'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.RemoteWithConnectionModelProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.ReturnsProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.SampleProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.SetConfigProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.SetProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.SettingsProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.SharingProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.SqlReadWriteProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.SqlSecurityProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.StabilityProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.StrictProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.TemporaryProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.TitleColumnConstraint'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.Timestamp'>: <function no_timestamp_sql>, <class 'sqlglot.expressions.ToMap'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.ToTableProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.TransformModelProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.TransientProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.UppercaseColumnConstraint'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.UnloggedProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.VarMap'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.ViewAttributeProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.VolatileProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.WithJournalTableProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.WithOperator'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.AnyValue'>: <function rename_func.<locals>.<lambda>>, <class 'sqlglot.expressions.ApproxDistinct'>: <function Presto.Generator.<lambda>>, <class 'sqlglot.expressions.ApproxQuantile'>: <function rename_func.<locals>.<lambda>>, <class 'sqlglot.expressions.ArgMax'>: <function rename_func.<locals>.<lambda>>, <class 'sqlglot.expressions.ArgMin'>: <function rename_func.<locals>.<lambda>>, <class 'sqlglot.expressions.Array'>: <function Presto.Generator.<lambda>>, <class 'sqlglot.expressions.ArrayAny'>: <function rename_func.<locals>.<lambda>>, <class 'sqlglot.expressions.ArrayConcat'>: <function rename_func.<locals>.<lambda>>, <class 'sqlglot.expressions.ArrayContains'>: <function rename_func.<locals>.<lambda>>, <class 'sqlglot.expressions.ArraySize'>: <function rename_func.<locals>.<lambda>>, <class 'sqlglot.expressions.ArrayToString'>: <function rename_func.<locals>.<lambda>>, <class 'sqlglot.expressions.ArrayUniqueAgg'>: <function rename_func.<locals>.<lambda>>, <class 'sqlglot.expressions.AtTimeZone'>: <function rename_func.<locals>.<lambda>>, <class 'sqlglot.expressions.BitwiseAnd'>: <function Presto.Generator.<lambda>>, <class 'sqlglot.expressions.BitwiseLeftShift'>: <function Presto.Generator.<lambda>>, <class 'sqlglot.expressions.BitwiseNot'>: <function Presto.Generator.<lambda>>, <class 'sqlglot.expressions.BitwiseOr'>: <function Presto.Generator.<lambda>>, <class 'sqlglot.expressions.BitwiseRightShift'>: <function Presto.Generator.<lambda>>, <class 'sqlglot.expressions.BitwiseXor'>: <function Presto.Generator.<lambda>>, <class 'sqlglot.expressions.Cast'>: <function preprocess.<locals>._to_sql>, <class 'sqlglot.expressions.CurrentTimestamp'>: <function Presto.Generator.<lambda>>, <class 'sqlglot.expressions.DateAdd'>: <function Presto.Generator.<lambda>>, <class 'sqlglot.expressions.DateDiff'>: <function Presto.Generator.<lambda>>, <class 'sqlglot.expressions.DateStrToDate'>: <function datestrtodate_sql>, <class 'sqlglot.expressions.DateToDi'>: <function Presto.Generator.<lambda>>, <class 'sqlglot.expressions.DateSub'>: <function Presto.Generator.<lambda>>, <class 'sqlglot.expressions.Decode'>: <function Presto.Generator.<lambda>>, <class 'sqlglot.expressions.DiToDate'>: <function Presto.Generator.<lambda>>, <class 'sqlglot.expressions.Encode'>: <function Presto.Generator.<lambda>>, <class 'sqlglot.expressions.FileFormatProperty'>: <function Presto.Generator.<lambda>>, <class 'sqlglot.expressions.First'>: <function _first_last_sql>, <class 'sqlglot.expressions.FirstValue'>: <function _first_last_sql>, <class 'sqlglot.expressions.FromTimeZone'>: <function Presto.Generator.<lambda>>, <class 'sqlglot.expressions.Group'>: <function preprocess.<locals>._to_sql>, <class 'sqlglot.expressions.GroupConcat'>: <function Presto.Generator.<lambda>>, <class 'sqlglot.expressions.If'>: <function if_sql.<locals>._if_sql>, <class 'sqlglot.expressions.ILike'>: <function no_ilike_sql>, <class 'sqlglot.expressions.Initcap'>: <function _initcap_sql>, <class 'sqlglot.expressions.ParseJSON'>: <function rename_func.<locals>.<lambda>>, <class 'sqlglot.expressions.Last'>: <function _first_last_sql>, <class 'sqlglot.expressions.LastValue'>: <function _first_last_sql>, <class 'sqlglot.expressions.LastDay'>: <function Presto.Generator.<lambda>>, <class 'sqlglot.expressions.Lateral'>: <function _explode_to_unnest_sql>, <class 'sqlglot.expressions.Left'>: <function left_to_substring_sql>, <class 'sqlglot.expressions.Levenshtein'>: <function rename_func.<locals>.<lambda>>, <class 'sqlglot.expressions.LogicalAnd'>: <function rename_func.<locals>.<lambda>>, <class 'sqlglot.expressions.LogicalOr'>: <function rename_func.<locals>.<lambda>>, <class 'sqlglot.expressions.Pivot'>: <function no_pivot_sql>, <class 'sqlglot.expressions.Quantile'>: <function _quantile_sql>, <class 'sqlglot.expressions.RegexpExtract'>: <function regexp_extract_sql>, <class 'sqlglot.expressions.Right'>: <function right_to_substring_sql>, <class 'sqlglot.expressions.SafeDivide'>: <function no_safe_divide_sql>, <class 'sqlglot.expressions.Schema'>: <function _schema_sql>, <class 'sqlglot.expressions.SchemaCommentProperty'>: <function Presto.Generator.<lambda>>, <class 'sqlglot.expressions.Select'>: <function preprocess.<locals>._to_sql>, <class 'sqlglot.expressions.SortArray'>: <function _no_sort_array>, <class 'sqlglot.expressions.StrPosition'>: <function Presto.Generator.<lambda>>, <class 'sqlglot.expressions.StrToDate'>: <function Presto.Generator.<lambda>>, <class 'sqlglot.expressions.StrToMap'>: <function rename_func.<locals>.<lambda>>, <class 'sqlglot.expressions.StrToTime'>: <function _str_to_time_sql>, <class 'sqlglot.expressions.StructExtract'>: <function struct_extract_sql>, <class 'sqlglot.expressions.Table'>: <function preprocess.<locals>._to_sql>, <class 'sqlglot.expressions.TimestampTrunc'>: <function timestamptrunc_sql.<locals>._timestamptrunc_sql>, <class 'sqlglot.expressions.TimeStrToDate'>: <function timestrtotime_sql>, <class 'sqlglot.expressions.TimeStrToTime'>: <function timestrtotime_sql>, <class 'sqlglot.expressions.TimeStrToUnix'>: <function Presto.Generator.<lambda>>, <class 'sqlglot.expressions.TimeToStr'>: <function Presto.Generator.<lambda>>, <class 'sqlglot.expressions.TimeToUnix'>: <function rename_func.<locals>.<lambda>>, <class 'sqlglot.expressions.ToChar'>: <function Presto.Generator.<lambda>>, <class 'sqlglot.expressions.TryCast'>: <function preprocess.<locals>._to_sql>, <class 'sqlglot.expressions.TsOrDiToDi'>: <function Presto.Generator.<lambda>>, <class 'sqlglot.expressions.TsOrDsAdd'>: <function _ts_or_ds_add_sql>, <class 'sqlglot.expressions.TsOrDsDiff'>: <function _ts_or_ds_diff_sql>, <class 'sqlglot.expressions.TsOrDsToDate'>: <function _ts_or_ds_to_date_sql>, <class 'sqlglot.expressions.Unhex'>: <function rename_func.<locals>.<lambda>>, <class 'sqlglot.expressions.UnixToStr'>: <function Presto.Generator.<lambda>>, <class 'sqlglot.expressions.UnixToTime'>: <function _unix_to_time_sql>, <class 'sqlglot.expressions.UnixToTimeStr'>: <function Presto.Generator.<lambda>>, <class 'sqlglot.expressions.VariancePop'>: <function rename_func.<locals>.<lambda>>, <class 'sqlglot.expressions.With'>: <function preprocess.<locals>._to_sql>, <class 'sqlglot.expressions.WithinGroup'>: <function preprocess.<locals>._to_sql>, <class 'sqlglot.expressions.Xor'>: <function bool_xor_sql>, <class 'sqlglot.expressions.MD5'>: <function Presto.Generator.<lambda>>, <class 'sqlglot.expressions.MD5Digest'>: <function rename_func.<locals>.<lambda>>, <class 'sqlglot.expressions.SHA'>: <function rename_func.<locals>.<lambda>>, <class 'sqlglot.expressions.SHA2'>: <function Presto.Generator.<lambda>>}
RESERVED_KEYWORDS = {'else', 'cross', 'except', 'intersect', 'current_time', 'alter', 'false', 'distinct', 'escape', 'extract', 'on', 'end', 'having', 'not', 'group', 'then', 'join', 'into', 'for', 'constraint', 'describe', 'right', 'insert', 'true', 'case', 'natural', 'left', 'exists', 'union', 'in', 'values', 'with', 'as', 'drop', 'select', 'where', 'is', 'null', 'execute', 'from', 'deallocate', 'delete', 'create', 'order', 'by', 'and', 'like', 'outer', 'or', 'prepare', 'using', 'between', 'when', 'current_timestamp', 'cast', 'table', 'inner', 'full'}
def strtounix_sql(self, expression: sqlglot.expressions.StrToUnix) -> str: View Source

    522        def strtounix_sql(self, expression: exp.StrToUnix) -> str:
    523            # Since `TO_UNIXTIME` requires a `TIMESTAMP`, we need to parse the argument into one.
    524            # To do this, we first try to `DATE_PARSE` it, but since this can fail when there's a
    525            # timezone involved, we wrap it in a `TRY` call and use `PARSE_DATETIME` as a fallback,
    526            # which seems to be using the same time mapping as Hive, as per:
    527            # https://joda-time.sourceforge.net/apidocs/org/joda/time/format/DateTimeFormat.html
    528            value_as_text = exp.cast(expression.this, exp.DataType.Type.TEXT)
    529            parse_without_tz = self.func("DATE_PARSE", value_as_text, self.format_time(expression))
    530            parse_with_tz = self.func(
    531                "PARSE_DATETIME",
    532                value_as_text,
    533                self.format_time(expression, Hive.INVERSE_TIME_MAPPING, Hive.INVERSE_TIME_TRIE),
    534            )
    535            coalesced = self.func("COALESCE", self.func("TRY", parse_without_tz), parse_with_tz)
    536            return self.func("TO_UNIXTIME", coalesced)

def bracket_sql(self, expression: sqlglot.expressions.Bracket) -> str: View Source

    538        def bracket_sql(self, expression: exp.Bracket) -> str:
    539            if expression.args.get("safe"):
    540                return self.func(
    541                    "ELEMENT_AT",
    542                    expression.this,
    543                    seq_get(
    544                        apply_index_offset(
    545                            expression.this,
    546                            expression.expressions,
    547                            1 - expression.args.get("offset", 0),
    548                        ),
    549                        0,
    550                    ),
    551                )
    552            return super().bracket_sql(expression)

def struct_sql(self, expression: sqlglot.expressions.Struct) -> str: View Source

    554        def struct_sql(self, expression: exp.Struct) -> str:
    555            from sqlglot.optimizer.annotate_types import annotate_types
    556
    557            expression = annotate_types(expression)
    558            values: t.List[str] = []
    559            schema: t.List[str] = []
    560            unknown_type = False
    561
    562            for e in expression.expressions:
    563                if isinstance(e, exp.PropertyEQ):
    564                    if e.type and e.type.is_type(exp.DataType.Type.UNKNOWN):
    565                        unknown_type = True
    566                    else:
    567                        schema.append(f"{self.sql(e, 'this')} {self.sql(e.type)}")
    568                    values.append(self.sql(e, "expression"))
    569                else:
    570                    values.append(self.sql(e))
    571
    572            size = len(expression.expressions)
    573
    574            if not size or len(schema) != size:
    575                if unknown_type:
    576                    self.unsupported(
    577                        "Cannot convert untyped key-value definitions (try annotate_types)."
    578                    )
    579                return self.func("ROW", *values)
    580            return f"CAST(ROW({', '.join(values)}) AS ROW({', '.join(schema)}))"

def interval_sql(self, expression: sqlglot.expressions.Interval) -> str: View Source

    582        def interval_sql(self, expression: exp.Interval) -> str:
    583            if expression.this and expression.text("unit").upper().startswith("WEEK"):
    584                return f"({expression.this.name} * INTERVAL '7' DAY)"
    585            return super().interval_sql(expression)

def transaction_sql(self, expression: sqlglot.expressions.Transaction) -> str: View Source

    587        def transaction_sql(self, expression: exp.Transaction) -> str:
    588            modes = expression.args.get("modes")
    589            modes = f" {', '.join(modes)}" if modes else ""
    590            return f"START TRANSACTION{modes}"

def generateseries_sql(self, expression: sqlglot.expressions.GenerateSeries) -> str: View Source

    592        def generateseries_sql(self, expression: exp.GenerateSeries) -> str:
    593            start = expression.args["start"]
    594            end = expression.args["end"]
    595            step = expression.args.get("step")
    596
    597            if isinstance(start, exp.Cast):
    598                target_type = start.to
    599            elif isinstance(end, exp.Cast):
    600                target_type = end.to
    601            else:
    602                target_type = None
    603
    604            if target_type and target_type.is_type("timestamp"):
    605                if target_type is start.to:
    606                    end = exp.cast(end, target_type)
    607                else:
    608                    start = exp.cast(start, target_type)
    609
    610            return self.func("SEQUENCE", start, end, step)

def offset_limit_modifiers( self, expression: sqlglot.expressions.Expression, fetch: bool, limit: Union[sqlglot.expressions.Fetch, sqlglot.expressions.Limit, NoneType]) -> List[str]: View Source

    612        def offset_limit_modifiers(
    613            self, expression: exp.Expression, fetch: bool, limit: t.Optional[exp.Fetch | exp.Limit]
    614        ) -> t.List[str]:
    615            return [
    616                self.sql(expression, "offset"),
    617                self.sql(limit),
    618            ]

def create_sql(self, expression: sqlglot.expressions.Create) -> str: View Source

    620        def create_sql(self, expression: exp.Create) -> str:
    621            """
    622            Presto doesn't support CREATE VIEW with expressions (ex: `CREATE VIEW x (cola)` then `(cola)` is the expression),
    623            so we need to remove them
    624            """
    625            kind = expression.args["kind"]
    626            schema = expression.this
    627            if kind == "VIEW" and schema.expressions:
    628                expression.this.set("expressions", None)
    629            return super().create_sql(expression)

Presto doesn't support CREATE VIEW with expressions (ex: `CREATE VIEW x (cola)` then `(cola)` is the expression), so we need to remove them

def delete_sql(self, expression: sqlglot.expressions.Delete) -> str: View Source

    631        def delete_sql(self, expression: exp.Delete) -> str:
    632            """
    633            Presto only supports DELETE FROM for a single table without an alias, so we need
    634            to remove the unnecessary parts. If the original DELETE statement contains more
    635            than one table to be deleted, we can't safely map it 1-1 to a Presto statement.
    636            """
    637            tables = expression.args.get("tables") or [expression.this]
    638            if len(tables) > 1:
    639                return super().delete_sql(expression)
    640
    641            table = tables[0]
    642            expression.set("this", table)
    643            expression.set("tables", None)
    644
    645            if isinstance(table, exp.Table):
    646                table_alias = table.args.get("alias")
    647                if table_alias:
    648                    table_alias.pop()
    649                    expression = t.cast(exp.Delete, expression.transform(unqualify_columns))
    650
    651            return super().delete_sql(expression)

Presto only supports DELETE FROM for a single table without an alias, so we need to remove the unnecessary parts. If the original DELETE statement contains more than one table to be deleted, we can't safely map it 1-1 to a Presto statement.

SELECT_KINDS: Tuple[str, ...] = ()
AFTER_HAVING_MODIFIER_TRANSFORMS = {'qualify': <function Generator.<lambda>>, 'windows': <function Generator.<lambda>>}

##### Inherited Members

sqlglot.generator.Generator
  Generator
  NULL_ORDERING_SUPPORTED
  IGNORE_NULLS_IN_FUNC
  LOCKING_READS_SUPPORTED
  EXPLICIT_UNION
  WRAP_DERIVED_VALUES
  CREATE_FUNCTION_RETURN_AS
  MATCHED_BY_SOURCE
  SINGLE_STRING_INTERVAL
  LIMIT_FETCH
  RENAME_TABLE_WITH_DB
  GROUPINGS_SEP
  INDEX_ON
  QUERY_HINT_SEP
  DUPLICATE_KEY_UPDATE_WITH_SET
  LIMIT_IS_TOP
  RETURNING_END
  COLUMN_JOIN_MARKS_SUPPORTED
  EXTRACT_ALLOWS_QUOTES
  VALUES_AS_TABLE
  ALTER_TABLE_INCLUDE_COLUMN_KEYWORD
  UNNEST_WITH_ORDINALITY
  AGGREGATE_FILTER_SUPPORTED
  SEMI_ANTI_JOIN_WITH_SIDE
  COMPUTED_COLUMN_WITH_TYPE
  SUPPORTS_TABLE_COPY
  TABLESAMPLE_REQUIRES_PARENS
  TABLESAMPLE_SIZE_IS_ROWS
  TABLESAMPLE_KEYWORDS
  TABLESAMPLE_WITH_METHOD
  TABLESAMPLE_SEED_KEYWORD
  COLLATE_IS_FUNC
  DATA_TYPE_SPECIFIERS_ALLOWED
  ENSURE_BOOLS
  CTE_RECURSIVE_KEYWORD_REQUIRED
  LAST_DAY_SUPPORTS_DATE_PART
  SUPPORTS_TABLE_ALIAS_COLUMNS
  UNPIVOT_ALIASES_ARE_IDENTIFIERS
  JSON_KEY_VALUE_PAIR_SEP
  INSERT_OVERWRITE
  SUPPORTS_SELECT_INTO
  SUPPORTS_UNLOGGED_TABLES
  SUPPORTS_CREATE_TABLE_LIKE
  JSON_TYPE_REQUIRED_FOR_EXTRACTION
  JSON_PATH_BRACKETED_KEY_SUPPORTED
  JSON_PATH_SINGLE_QUOTE_ESCAPE
  SUPPORTED_JSON_PATH_PARTS
  CAN_IMPLEMENT_ARRAY_ANY
  OUTER_UNION_MODIFIERS
  COPY_PARAMS_ARE_WRAPPED
  COPY_PARAMS_EQ_REQUIRED
  COPY_HAS_INTO_KEYWORD
  TRY_SUPPORTED
  STAR_EXCEPT
  WITH_PROPERTIES_PREFIX
  TIME_PART_SINGULARS
  TOKEN_MAPPING
  PARAMETER_TOKEN
  NAMED_PLACEHOLDER_TOKEN
  WITH_SEPARATED_COMMENTS
  EXCLUDE_COMMENTS
  UNWRAPPED_INTERVAL_VALUES
  PARAMETERIZABLE_TEXT_TYPES
  EXPRESSIONS_WITHOUT_NESTED_CTES
  SENTINEL_LINE_BREAK
  pretty
  identify
  normalize
  pad
  unsupported_level
  max_unsupported
  leading_comma
  max_text_width
  comments
  dialect
  normalize_functions
  unsupported_messages
  generate
  preprocess
  unsupported
  sep
  seg
  pad_comment
  maybe_comment
  wrap
  no_identify
  normalize_func
  indent
  sql
  uncache_sql
  cache_sql
  characterset_sql
  column_parts
  column_sql
  columnposition_sql
  columndef_sql
  columnconstraint_sql
  computedcolumnconstraint_sql
  autoincrementcolumnconstraint_sql
  compresscolumnconstraint_sql
  generatedasidentitycolumnconstraint_sql
  generatedasrowcolumnconstraint_sql
  periodforsystemtimeconstraint_sql
  notnullcolumnconstraint_sql
  transformcolumnconstraint_sql
  primarykeycolumnconstraint_sql
  uniquecolumnconstraint_sql
  createable_sql
  sequenceproperties_sql
  clone_sql
  describe_sql
  heredoc_sql
  prepend_ctes
  with_sql
  cte_sql
  tablealias_sql
  bitstring_sql
  hexstring_sql
  bytestring_sql
  unicodestring_sql
  rawstring_sql
  datatypeparam_sql
  datatype_sql
  directory_sql
  drop_sql
  except_sql
  except_op
  fetch_sql
  filter_sql
  hint_sql
  indexparameters_sql
  index_sql
  identifier_sql
  hex_sql
  lowerhex_sql
  inputoutputformat_sql
  national_sql
  partition_sql
  properties_sql
  root_properties
  properties
  with_properties
  locate_properties
  property_name
  property_sql
  likeproperty_sql
  fallbackproperty_sql
  journalproperty_sql
  freespaceproperty_sql
  checksumproperty_sql
  mergeblockratioproperty_sql
  datablocksizeproperty_sql
  blockcompressionproperty_sql
  isolatedloadingproperty_sql
  partitionboundspec_sql
  partitionedofproperty_sql
  lockingproperty_sql
  withdataproperty_sql
  withsystemversioningproperty_sql
  insert_sql
  intersect_sql
  intersect_op
  introducer_sql
  kill_sql
  pseudotype_sql
  objectidentifier_sql
  onconflict_sql
  returning_sql
  rowformatdelimitedproperty_sql
  withtablehint_sql
  indextablehint_sql
  historicaldata_sql
  table_parts
  table_sql
  tablesample_sql
  pivot_sql
  version_sql
  tuple_sql
  update_sql
  values_sql
  var_sql
  into_sql
  from_sql
  group_sql
  having_sql
  connect_sql
  prior_sql
  join_sql
  lambda_sql
  lateral_op
  lateral_sql
  limit_sql
  offset_sql
  setitem_sql
  set_sql
  pragma_sql
  lock_sql
  literal_sql
  escape_str
  loaddata_sql
  null_sql
  boolean_sql
  order_sql
  withfill_sql
  cluster_sql
  distribute_sql
  sort_sql
  ordered_sql
  matchrecognizemeasure_sql
  matchrecognize_sql
  query_modifiers
  queryoption_sql
  after_limit_modifiers
  select_sql
  schema_sql
  schema_columns_sql
  star_sql
  parameter_sql
  sessionparameter_sql
  placeholder_sql
  subquery_sql
  qualify_sql
  set_operations
  union_sql
  union_op
  unnest_sql
  prewhere_sql
  where_sql
  window_sql
  partition_by_sql
  windowspec_sql
  withingroup_sql
  between_sql
  bracket_offset_expressions
  all_sql
  any_sql
  exists_sql
  case_sql
  constraint_sql
  nextvaluefor_sql
  extract_sql
  trim_sql
  convert_concat_args
  concat_sql
  concatws_sql
  check_sql
  foreignkey_sql
  primarykey_sql
  if_sql
  matchagainst_sql
  jsonkeyvalue_sql
  jsonpath_sql
  json_path_part
  formatjson_sql
  jsonobject_sql
  jsonobjectagg_sql
  jsonarray_sql
  jsonarrayagg_sql
  jsoncolumndef_sql
  jsonschema_sql
  jsontable_sql
  openjsoncolumndef_sql
  openjson_sql
  in_sql
  in_unnest_op
  return_sql
  reference_sql
  anonymous_sql
  paren_sql
  neg_sql
  not_sql
  alias_sql
  pivotalias_sql
  aliases_sql
  atindex_sql
  attimezone_sql
  fromtimezone_sql
  add_sql
  and_sql
  or_sql
  xor_sql
  connector_sql
  bitwiseand_sql
  bitwiseleftshift_sql
  bitwisenot_sql
  bitwiseor_sql
  bitwiserightshift_sql
  bitwisexor_sql
  cast_sql
  currentdate_sql
  currenttimestamp_sql
  collate_sql
  command_sql
  comment_sql
  mergetreettlaction_sql
  mergetreettl_sql
  commit_sql
  rollback_sql
  altercolumn_sql
  alterdiststyle_sql
  altersortkey_sql
  renametable_sql
  renamecolumn_sql
  alterset_sql
  altertable_sql
  add_column_sql
  droppartition_sql
  addconstraint_sql
  distinct_sql
  ignorenulls_sql
  respectnulls_sql
  havingmax_sql
  intdiv_sql
  dpipe_sql
  div_sql
  overlaps_sql
  distance_sql
  dot_sql
  eq_sql
  propertyeq_sql
  escape_sql
  glob_sql
  gt_sql
  gte_sql
  ilike_sql
  ilikeany_sql
  is_sql
  like_sql
  likeany_sql
  similarto_sql
  lt_sql
  lte_sql
  mod_sql
  mul_sql
  neq_sql
  nullsafeeq_sql
  nullsafeneq_sql
  slice_sql
  sub_sql
  trycast_sql
  try_sql
  log_sql
  use_sql
  binary
  function_fallback_sql
  func
  format_args
  too_wide
  format_time
  expressions
  op_expressions
  naked_property
  tag_sql
  token_sql
  userdefinedfunction_sql
  joinhint_sql
  kwarg_sql
  when_sql
  merge_sql
  tochar_sql
  tonumber_sql
  dictproperty_sql
  dictrange_sql
  dictsubproperty_sql
  oncluster_sql
  clusteredbyproperty_sql
  anyvalue_sql
  querytransform_sql
  indexconstraintoption_sql
  checkcolumnconstraint_sql
  indexcolumnconstraint_sql
  nvl2_sql
  comprehension_sql
  columnprefix_sql
  opclass_sql
  predict_sql
  forin_sql
  refresh_sql
  operator_sql
  toarray_sql
  tsordstotime_sql
  tsordstotimestamp_sql
  tsordstodate_sql
  unixdate_sql
  lastday_sql
  dateadd_sql
  arrayany_sql
  partitionrange_sql
  truncatetable_sql
  convert_sql
  copyparameter_sql
  credentials_sql
  copy_sql
  semicolon_sql
  datadeletionproperty_sql